"""
Defines the tree widget to display the contents of a selected directory.
"""

import re
import os
import sys
import subprocess
import fnmatch
import os.path as op

import pyzo
from pyzo import translate
from . import QtCore, QtGui, QtWidgets

from . import tasks
from .utils import hasHiddenAttribute, getMounts, cleanpath, isdir, ext


# How to name the list of drives/mounts (i.e. 'my computer')
MOUNTS = "drives"


# Create icon provider
iconprovider = QtWidgets.QFileIconProvider()


def addIconOverlays(icon, *overlays, offset=(8, 0), overlay_offset=(0, 0)):
    """Create an overlay for an icon."""
    # Create painter and pixmap
    pm0 = QtGui.QPixmap(16 + offset[0], 16)  # icon.pixmap(16+offset[0],16+offset[1])
    pm0.fill(QtGui.QColor(0, 0, 0, 0))
    painter = QtGui.QPainter()
    painter.begin(pm0)
    # Draw original icon
    painter.drawPixmap(offset[0], offset[1], icon.pixmap(16, 16))
    # Draw overlays
    for overlay in overlays:
        pm1 = overlay.pixmap(16, 16)
        painter.drawPixmap(overlay_offset[0], overlay_offset[1], pm1)
    # Finish
    painter.end()
    # Done (return resulting icon)
    return QtGui.QIcon(pm0)


def _filterFileByName(path, filters, base_path):
    # Init default; return True if there are no filters
    default = True
    matchCase = False

    for filter in filters:
        # Process filters in order
        if "/" in filter:
            filename = op.relpath(path, base_path).replace("\\", "/")
            filename = "/" + filename.lstrip("/")
        else:
            filename = op.basename(path)
        if not matchCase:
            filter = filter.lower()
            filename = filename.lower()
        if filter.startswith("!"):
            # If the filename matches an exclusive filter, hide it
            if fnmatch.fnmatch(filename, filter[1:]):
                return False
            default = True
        else:
            # If the file name matches a filter not starting with!, show it
            if fnmatch.fnmatch(filename, filter):
                return True
            default = False

    return default


def createMounts(browser, tree):
    """Create items for all known mount points (i.e. drives on Windows)."""
    fsProxy = browser._fsProxy

    mountPoints = getMounts()
    mountPoints.sort(key=lambda x: x.lower())
    for entry in mountPoints:
        entry = cleanpath(entry)
        DriveItem(tree, fsProxy.dir(entry))


def createItemsFun(browser, parent):
    """Create the tree widget items for a Tree or DirItem."""

    # Get file system proxy and dir proxy for which we shall create items
    fsProxy = browser._fsProxy
    dirProxy = parent._proxy

    # Get meta information from browser
    searchFilter = browser.searchFilter()
    searchFilter = searchFilter if searchFilter["pattern"] else None
    expandedDirs = browser.expandedDirs
    starredDirs = browser.starredDirs

    if searchFilter and searchFilter["regExp"]:
        try:
            re.compile(searchFilter["pattern"], re.MULTILINE)
        except re.error as err:
            ErrorItem(parent, "Error in regular expression:\n{}".format(err))
            return

    # Prepare name filter info
    nameFilters = browser.nameFilter().replace(",", " ").split()
    hideHidden = "!hidden" in nameFilters
    nameFilters = [f for f in nameFilters if f not in ("", "!hiddden", "hidden")]

    # Filter the contents of this folder
    try:
        dirs = []
        for entry in dirProxy.dirs():
            entry = cleanpath(entry)
            if hideHidden:
                if op.basename(entry).startswith("."):
                    continue  # Skip hidden files
                if hasHiddenAttribute(entry):
                    continue  # Skip hidden files on Windows
                if op.basename(entry) == "node_modules":
                    continue
            if op.basename(entry) == "__pycache__":
                continue
            dirs.append(entry)

        files = []
        for entry in dirProxy.files():
            entry = cleanpath(entry)
            if hideHidden and op.basename(entry).startswith("."):
                continue  # Skip hidden files
            if hideHidden and hasHiddenAttribute(entry):
                continue  # Skip hidden files on Windows
            if not _filterFileByName(entry, nameFilters, browser._tree.path()):
                continue
            files.append(entry)

    except (OSError, IOError) as err:
        ErrorItem(parent, str(err))
        return

    # Sort dirs (case insensitive, number aware)
    dirs.sort(key=dirpathSplitName2sortkey)

    # Sort files (case insensitive, first by type, then by name, number aware)
    files.sort(key=filepathSplitName2sortkey)

    if not searchFilter:
        # Create dirs
        for path in dirs:
            starred = op.normcase(path) in starredDirs
            item = DirItem(parent, fsProxy.dir(path), starred)
            # Set hidden, we can safely expand programmatically when hidden
            item.setHidden(True)
            # Set expanded and visibility
            if op.normcase(path) in expandedDirs:
                item.setExpanded(True)
            item.setHidden(False)

        # Create files
        for path in files:
            item = FileItem(parent, fsProxy.file(path))

    else:
        # If searching, inject everything in the tree
        # And every item is hidden at first
        parent = browser._tree
        if parent.topLevelItemCount():
            searchInfoItem = parent.topLevelItem(0)
        else:
            searchInfoItem = SearchInfoItem(parent)

        # Increase number of found files
        searchInfoItem.increaseTotal(len(files))

        # Create temporary file items
        for path in files:
            item = TemporaryFileItem(parent, fsProxy.file(path))
            item.search(searchFilter)

        # Create temporary dir items
        if searchFilter["subDirs"]:
            for path in dirs:
                if os.path.basename(path) not in (".git", ".hg"):
                    item = TemporaryDirItem(parent, fsProxy.dir(path))

    # Return number of files added
    return len(dirs) + len(files)


def filename2sortkey(s):
    """converts a filename string to a tuple of strings and ints which can
    be used as a key for sorting

    example:
    'image009_part4.png' --> ('image', 9, '_part', 4, '.png')
    """
    parts = re.split(r"(\d+)", s)
    return tuple(int(s) if s.isdecimal() else s for s in parts)


def filepathSplitName2sortkey(filepath):
    """converts the filename of a filepath to a tuple that can be used to
    logically sort them. Sorting first by extension and case-insensitive.
    """
    name = os.path.basename(filepath).lower()
    name, ext = os.path.splitext(name)
    return (ext,) + filename2sortkey(name)


def dirpathSplitName2sortkey(dirpath):
    """converts the dirname of a dirpath to a tuple that can be used to
    logically sort them. Sorting is case-insensitive.
    """
    name = os.path.basename(dirpath).lower()
    return filename2sortkey(name)


class BrowserItem(QtWidgets.QTreeWidgetItem):
    """Abstract item in the tree widget."""

    def __init__(self, parent, pathProxy, *args):
        self._proxy = pathProxy
        super().__init__(parent, [], *args)
        # Set pathname to show, and icon
        strippedParentPath = parent.path().rstrip("/\\")
        if self.path().startswith(strippedParentPath):
            basename = self.path()[len(strippedParentPath) + 1 :]
        else:
            basename = self.path()  #  For mount points
        self.setText(0, basename)
        self.setFileIcon()
        # Setup interface with proxy
        self._proxy.changed.connect(self.onChanged)
        self._proxy.deleted.connect(self.onDeleted)
        self._proxy.errored.connect(self.onErrored)
        self._proxy.taskFinished.connect(self.onTaskFinished)

    def path(self):
        return self._proxy.path()

    def _createDummyItem(self, txt):
        ErrorItem(self, txt)
        # QtWidgets.QTreeWidgetItem(self, [txt])

    def onDestroyed(self):
        self._proxy.cancel()

    def clear(self):
        """Clear method that calls onDestroyed on its children."""
        for i in reversed(range(self.childCount())):
            item = self.child(i)
            if hasattr(item, "onDestroyed"):
                item.onDestroyed()
            self.removeChild(item)

    # To overload ...

    def onChanged(self):
        pass

    def onDeleted(self):
        pass

    def onErrored(self, err):
        self.clear()
        self._createDummyItem("Error: " + err)

    def onTaskFinished(self, task):
        # Getting the result raises exception if an error occured.
        # Which is what we want; so it is visible in the logger shell
        task.result()


class DriveItem(BrowserItem):
    """Tree widget item for directories."""

    def __init__(self, parent, pathProxy):
        super().__init__(parent, pathProxy)
        # Item is not expandable

    def setFileIcon(self):
        # Use folder icon
        self.setIcon(0, pyzo.icons.drive)

    def onActivated(self):
        self.treeWidget().setPath(self.path())


class DirItem(BrowserItem):
    """Tree widget item for directories."""

    def __init__(self, parent, pathProxy, starred=False):
        self._starred = starred
        super().__init__(parent, pathProxy)

        # Create dummy item so that the dir is expandable
        self._createDummyItem("Loading contents ...")

    def setFileIcon(self):
        # Use folder icon
        icon = iconprovider.icon(iconprovider.IconType.Folder)
        overlays = []
        if self._starred:
            overlays.append(pyzo.icons.bullet_yellow)
        icon = addIconOverlays(icon, *overlays, offset=(8, 0), overlay_offset=(-4, 0))
        self.setIcon(0, icon)

    def onActivated(self):
        self.treeWidget().setPath(self.path())

    def onExpanded(self):
        # Update list of expanded dirs
        expandedDirs = self.treeWidget().parent().expandedDirs
        p = op.normcase(self.path())  # Normalize case!
        if p not in expandedDirs:
            expandedDirs.append(p)
        # Keep track of changes in our contents
        self._proxy.track()
        self._proxy.push()

    def onCollapsed(self):
        # Update list of expanded dirs
        expandedDirs = self.treeWidget().parent().expandedDirs
        p = op.normcase(self.path())  # Normalize case!
        while p in expandedDirs:
            expandedDirs.remove(p)
        # Stop tracking changes in our contents
        self._proxy.cancel()
        # Clear contents and create a single placeholder item
        self.clear()
        self._createDummyItem("Loading contents ...")

    # No need to implement onDeleted: the parent will get a changed event.

    def onChanged(self):
        """Called when a change in the contents has occured, or when
        we just activated the proxy. Update our items!
        """
        if not self.isExpanded():
            return
        tree = self.treeWidget()
        tree.createItems(self)


class FileItem(BrowserItem):
    """Tree widget item for files."""

    def __init__(self, parent, pathProxy, mode="normal"):
        super().__init__(parent, pathProxy)
        self._mode = mode
        self._timeSinceLastDocString = 0

        if self._mode == "normal" and self.path().lower().endswith(".py"):
            self._createDummyItem("Loading high level structure ...")

    def setFileIcon(self):
        # Create dummy file in pyzo user dir
        dummy_filename = op.join(
            cleanpath(pyzo.appDataDir), "dummyFiles", "dummy" + ext(self.path())
        )
        # Create file?
        if not op.isfile(dummy_filename):
            if not isdir(op.dirname(dummy_filename)):
                os.makedirs(op.dirname(dummy_filename))
            with open(dummy_filename, "wb"):
                pass
        # Use that file
        if sys.platform.startswith("linux") and not QtCore.__file__.startswith("/usr/"):
            icon = iconprovider.icon(iconprovider.IconType.File)
        else:
            icon = iconprovider.icon(QtCore.QFileInfo(dummy_filename))
        icon = addIconOverlays(icon)
        self.setIcon(0, icon)

    def searchContents(self, needle, **kwargs):
        self.setHidden(True)
        self._proxy.setSearch(needle, **kwargs)

    def onActivated(self):
        # todo: someday we should be able to simply pass the proxy object to the editors
        # so that we can open files on any file system
        path = self.path()
        if ext(path) not in [".pyc", ".pyo", ".png", ".jpg", ".ico"]:
            # Load file
            pyzo.editors.loadFile(path)
            # Give focus
            pyzo.editors.getCurrentEditor().setFocus()

    def onExpanded(self):
        if self._mode == "normal":
            # Create task to retrieve high level structure
            if self.path().lower().endswith(".py"):
                self._proxy.pushTask(tasks.DocstringTask())
                self._proxy.pushTask(tasks.PeekTask())

    def onCollapsed(self):
        if self._mode == "normal":
            self.clear()
            if self.path().lower().endswith(".py"):
                self._createDummyItem("Loading high level structure ...")

    #     def onClicked(self):
    #         # Limit sending events to prevent flicker when double clicking
    #         if time.time() - self._timeSinceLastDocString < 0.5:
    #             return
    #         self._timeSinceLastDocString = time.time()
    #         # Create task
    #         if self.path().lower().endswith('.py'):
    #             self._proxy.pushTask(tasks.DocstringTask())

    def onChanged(self):
        pass

    def onTaskFinished(self, task):
        if isinstance(task, tasks.DocstringTask):
            result = task.result()
            self.clear()  # Docstring task is done *before* peek task
            if result:
                DocstringItem(self, result)
        #         if isinstance(task, tasks.DocstringTask):
        #             result = task.result()
        #             if result:
        #                 #self.setToolTip(0, result)
        #                 # Show tooltip *now* if mouse is still over this item
        #                 tree = self.treeWidget()
        #                 pos = tree.mapFromGlobal(QtGui.QCursor.pos())
        #                 if tree.itemAt(pos) is self:
        #                     QtWidgets.QToolTip.showText(QtGui.QCursor.pos(), result)
        elif isinstance(task, tasks.PeekTask):
            result = task.result()
            # self.clear()  # Cleared when docstring task result is received
            if result:
                for r in result:
                    SubFileItem(self, *r)
            else:
                self._createDummyItem("No classes or functions found.")
        else:
            super().onTaskFinished(task)


class SubFileItem(QtWidgets.QTreeWidgetItem):
    """Tree widget item for search items."""

    def __init__(self, parent, linenr, text, showlinenr=False):
        super().__init__(parent)
        self._linenr = linenr
        if showlinenr:
            self.setText(0, "Line {}: {}".format(linenr, text))
        else:
            self.setText(0, text)

    def path(self):
        return self.parent().path()

    def onActivated(self):
        path = self.path()
        if ext(path) not in [".pyc", ".pyo", ".png", ".jpg", ".ico"]:
            # Load and get editor
            fileItem = pyzo.editors.loadFile(path)
            editor = fileItem._editor
            # Goto line
            editor.gotoLine(self._linenr)
            # Give focus
            pyzo.editors.getCurrentEditor().setFocus()


class DocstringItem(QtWidgets.QTreeWidgetItem):
    """Tree widget item for docstring placeholder items."""

    def __init__(self, parent, docstring):
        super().__init__(parent)
        self._docstring = docstring
        # Get one-line version of docstring
        shortText = self._docstring.split("\n", 1)[0].strip()
        if len(shortText) < len(self._docstring):
            shortText += "..."
        # Set short version now
        self.setText(0, "doc: " + shortText)
        # Long version is the tooltip
        self.setToolTip(0, docstring)

    def path(self):
        return self.parent().path()

    def onClicked(self):
        tree = self.treeWidget()
        pos = tree.mapFromGlobal(QtGui.QCursor.pos())
        if tree.itemAt(pos) is self:
            QtWidgets.QToolTip.showText(QtGui.QCursor.pos(), self._docstring)


class ErrorItem(QtWidgets.QTreeWidgetItem):
    """Tree widget item for errors and information."""

    def __init__(self, parent, info):
        super().__init__(parent)
        self.setText(0, info)
        self.setFlags(QtCore.Qt.ItemFlag.NoItemFlags)
        font = self.font(0)
        font.setItalic(True)
        self.setFont(0, font)


class SearchInfoItem(ErrorItem):
    """Tree widget item that displays info on the search."""

    def __init__(self, parent):
        super().__init__(parent, "Searching ...")
        self._totalCount = 0
        self._checkCount = 0
        self._hitCount = 0

    def increaseTotal(self, c):
        self._totalCount += c
        self.updateCounts()

    def addFile(self, hit):
        self._checkCount += 1
        if hit:
            self._hitCount += 1
        # Update appearance
        self.updateCounts()

    def updateCounts(self):
        counts = self._checkCount, self._totalCount, self._hitCount
        self.setText(0, "Searched {}/{} files: {} hits".format(*counts))


class TemporaryDirItem:
    """Created when searching. This object posts a requests for its contents
    which are then processed, after which this object disbands itself.
    """

    __slots__ = ["__weakref__", "_proxy", "_tree"]

    def __init__(self, tree, pathProxy):
        self._tree = tree
        self._proxy = pathProxy
        self._proxy.changed.connect(self.onChanged)
        # Process asap, but do not track
        self._proxy.push()
        # Store ourself
        tree._temporaryItems.add(self)

    def clear(self):
        pass  # tree.createItems() calls this ...

    def onChanged(self):
        # Disband
        self._tree._temporaryItems.discard(self)
        # Process contents
        self._tree.createItems(self)


class TemporaryFileItem:
    """Created when searching. This object posts a requests to search
    its contents which are then processed, after which this object
    disbands itself, passing the proxy object to a real FileItem if the
    search had results.
    """

    __slots__ = ["__weakref__", "_proxy", "_tree"]

    def __init__(self, tree, pathProxy):
        self._tree = tree
        self._proxy = pathProxy
        self._proxy.taskFinished.connect(self.onSearchResult)
        # Store ourself
        tree._temporaryItems.add(self)

    def search(self, searchFilter):
        self._proxy.pushTask(tasks.SearchTask(**searchFilter))

    def onSearchResult(self, task):
        # Disband now
        self._tree._temporaryItems.discard(self)

        # Get result. May raise an error
        result = task.result()
        # Process contents
        if result:
            item = FileItem(self._tree, self._proxy, "search")  # Search mode
            for r in result:
                SubFileItem(item, *r, showlinenr=True)
        # Update counter
        searchInfoItem = self._tree.topLevelItem(0)
        if isinstance(searchInfoItem, SearchInfoItem):
            searchInfoItem.addFile(bool(result))


class Tree(QtWidgets.QTreeWidget):
    """Representation of the tree view.
    Instances of this class are responsible for keeping the contents
    up-to-date. The Item classes above are dumb objects.
    """

    dirChanged = QtCore.Signal(str)  # Emitted when user goes into a subdir

    def __init__(self, parent):
        super().__init__(parent)

        # Initialize
        self.setMinimumWidth(150)
        self.setMinimumHeight(150)
        #
        self.setColumnCount(1)
        self.setHeaderHidden(True)
        self.setIconSize(QtCore.QSize(24, 16))

        # Connect signals
        self.itemExpanded.connect(self.onItemExpanded)
        self.itemCollapsed.connect(self.onItemCollapsed)
        self.itemClicked.connect(self.onItemClicked)
        self.itemActivated.connect(self.onItemActivated)

        # Variables for restoring the view after updating
        self._selectedPath = ""  # To restore a selection after updating
        self._selectedScrolling = 0

        # Set of temporary items
        self._temporaryItems = set()

        # Define context menu
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.contextMenuTriggered)

        # Initialize proxy (this is where the path is stored)
        self._proxy = None

    def path(self):
        """Get the current path shown by the treeview."""
        return self._proxy.path()

    def setPath(self, path):
        """Set the current path shown by the treeview."""
        # Close old proxy
        if self._proxy is not None:
            self._proxy.cancel()
            failed = []
            for signal, func in [
                (self.destroyed, self._proxy.cancel),
                (self._proxy.changed, self.onChanged),
                (self._proxy.deleted, self.onDeleted),
                (self._proxy.errored, self.onErrored),
            ]:
                try:
                    signal.disconnect(func)
                except Exception:
                    failed.append(signal)
            if failed:
                pass  # print("Failed to disconnect", failed)
        # Create new proxy
        if True:
            self._proxy = self.parent()._fsProxy.dir(path)
            self._proxy.changed.connect(self.onChanged)
            self._proxy.deleted.connect(self.onDeleted)
            self._proxy.errored.connect(self.onErrored)
            self.destroyed.connect(self._proxy.cancel)
        # Activate the proxy, we'll get a call at onChanged() asap.
        if path.lower() == MOUNTS.lower():
            self.clear()
            createMounts(self.parent(), self)
        else:
            self._proxy.track()
            self._proxy.push()
        # Store dir in config
        self.parent().config.path = path
        # Signal that the dir has changed
        # Note that our contents may not be visible yet.
        self.dirChanged.emit(self.path())

    def setPathUp(self):
        """Go one directory up."""
        newPath = op.dirname(self.path())

        if op.normcase(newPath) == op.normcase(self.path()):
            self.setPath(cleanpath(MOUNTS))
        else:
            self.setPath(newPath)

    def clear(self):
        """Overload the clear method to remove the items in a nice
        way, allowing the pathProxy instance to be closed correctly.
        """
        # Clear temporary (invisible) items
        for item in self._temporaryItems:
            item._proxy.cancel()
        self._temporaryItems.clear()
        # Clear visible items
        for i in reversed(range(self.topLevelItemCount())):
            item = self.topLevelItem(i)
            if hasattr(item, "clear"):
                item.clear()
            if hasattr(item, "onDestroyed"):
                item.onDestroyed()
        super().clear()

    def mouseDoubleClickEvent(self, event):
        """Bypass expanding an item when double-clicking it.
        Only activate the item.
        """
        item = self.itemAt(event.position().toPoint())
        if item is not None:
            self.onItemActivated(item)

    def onChanged(self):
        """Called when our contents change or when we just changed directories."""
        self.createItems(self)

    def createItems(self, parent):
        """High level method to create the items of the tree or a DirItem.
        This method will handle the restoring of state etc.
        The actual filtering of entries and creation of tree widget items
        is done in the createItemsFun() function.
        """
        # Store state and clear
        self._storeSelectionState()
        parent.clear()
        # Create sub items
        count = createItemsFun(self.parent(), parent)
        if not count and isinstance(parent, QtWidgets.QTreeWidgetItem):
            ErrorItem(parent, "Empty / no matches")
        # Restore state
        self._restoreSelectionState()

    def onErrored(self, err="..."):
        self.clear()
        ErrorItem(self, "Error: " + err)

    def onDeleted(self):
        self.setPathUp()

    def onItemExpanded(self, item):
        if hasattr(item, "onExpanded"):
            item.onExpanded()

    def onItemCollapsed(self, item):
        if hasattr(item, "onCollapsed"):
            item.onCollapsed()

    def onItemClicked(self, item):
        if hasattr(item, "onClicked"):
            item.onClicked()

    def onItemActivated(self, item):
        """When an item is "activated", make that the new directory,
        or open that file.
        """
        if hasattr(item, "onActivated"):
            item.onActivated()

    def _storeSelectionState(self):
        # Store selection
        items = self.selectedItems()
        self._selectedPath = items[0].path() if items else ""
        # Store scrolling
        self._selectedScrolling = self.verticalScrollBar().value()

    def _restoreSelectionState(self):
        # First select the first item
        # (otherwise the scrolling won't work for some reason)
        if self.topLevelItemCount():
            self.setCurrentItem(self.topLevelItem(0))
        # Restore selection
        if self._selectedPath:
            items = self.findItems(
                op.basename(self._selectedPath), QtCore.Qt.MatchFlag.MatchExactly, 0
            )
            items = [
                item
                for item in items
                if op.normcase(item.path()) == op.normcase(self._selectedPath)
            ]
            if items:
                self.setCurrentItem(items[0])
        # Restore scrolling
        self.verticalScrollBar().setValue(self._selectedScrolling)
        self.verticalScrollBar().setValue(self._selectedScrolling)

    def contextMenuTriggered(self, p):
        """Called when context menu is clicked"""
        # Get item that was clicked on
        item = self.itemAt(p)
        if item is None:
            item = self

        # Create and show menu
        if isinstance(item, (Tree, FileItem, DirItem)):
            menu = PopupMenu(self, item)
            menu.popup(self.mapToGlobal(p + QtCore.QPoint(3, 3)))


class PopupMenu(pyzo.core.menu.Menu):
    def __init__(self, parent, item):
        self._item = item
        super().__init__(parent, " ")

    def build(self):
        isplat = sys.platform.startswith

        # The star object
        if isinstance(self._item, DirItem):
            if self._item._starred:
                self.addItem(
                    translate("filebrowser", "Unstar this directory"), None, self._star
                )
            else:
                self.addItem(
                    translate("filebrowser", "Star this directory"), None, self._star
                )
            self.addSeparator()

        # Pyzo related functions
        if isinstance(self._item, FileItem):
            self.addItem(translate("filebrowser", "Open"), None, self._item.onActivated)
            if self._item.path().endswith(".py"):
                self.addItem(
                    translate("filebrowser", "Run as script"), None, self._runAsScript
                )
            elif self._item.path().endswith(".ipynb"):
                self.addItem(
                    translate("filebrowser", "Run Jupyter notebook"),
                    None,
                    self._runNotebook,
                )
            else:
                self.addItem(
                    translate("filebrowser", "Import data..."), None, self._importData
                )
            self.addSeparator()

        # Create items for open and copy path
        if isinstance(self._item, (FileItem, DirItem)):
            if isplat("win") or isplat("darwin") or isplat("linux"):
                self.addItem(
                    translate("filebrowser", "Open outside Pyzo"),
                    None,
                    self._openOutsidePyzo,
                )
            if isplat("darwin"):
                self.addItem(
                    translate("filebrowser", "Reveal in Finder"),
                    None,
                    self._showInFinder,
                )
            if True:
                self.addItem(
                    translate("filebrowser", "Copy path"), None, self._copyPath
                )
            self.addSeparator()

        # Create items for file management
        if isinstance(self._item, FileItem):
            self.addItem(translate("filebrowser", "Rename"), None, self.onRename)
            self.addItem(translate("filebrowser", "Delete"), None, self.onDelete)
            # self.addItem(translate("filebrowser", "Duplicate"), None, self.onDuplicate)
        if isinstance(self._item, (Tree, DirItem)):
            self.addItem(
                translate("filebrowser", "Create new file"), None, self.onCreateFile
            )
            self.addItem(
                translate("filebrowser", "Create new directory"), None, self.onCreateDir
            )
        if isinstance(self._item, DirItem):
            self.addSeparator()
            self.addItem(translate("filebrowser", "Rename"), None, self.onRename)
            self.addItem(translate("filebrowser", "Delete"), None, self.onDelete)

    def _star(self):
        # Prepare
        browser = self.parent().parent()
        path = self._item.path()
        if self._item._starred:
            browser.removeStarredDir(path)
        else:
            browser.addStarredDir(path)
        # Refresh
        self.parent().setPath(self.parent().path())

    def _openOutsidePyzo(self):
        path = self._item.path()
        if sys.platform.startswith("darwin"):
            subprocess.call(("open", path))
        elif sys.platform.startswith("win"):
            if " " in path:  # http://stackoverflow.com/a/72796/2271927
                subprocess.call(("start", "", path), shell=True)
            else:
                subprocess.call(("start", path), shell=True)
        elif sys.platform.startswith("linux"):
            # xdg-open is available on all Freedesktop.org compliant distros
            # http://superuser.com/questions/38984/linux-equivalent-command-for-open-command-on-mac-windows
            subprocess.call(("xdg-open", path))

    def _showInFinder(self):
        subprocess.call(("open", "-R", self._item.path()))

    def _copyPath(self):
        QtWidgets.qApp.clipboard().setText(self._item.path())

    def _runAsScript(self):
        filename = self._item.path()
        shell = pyzo.shells.getCurrentShell()
        if shell is not None:
            shell.restart(filename)
        else:
            msg = "No shell to run code in. "
            m = QtWidgets.QMessageBox(self)
            m.setWindowTitle(translate("menu dialog", "Could not run"))
            m.setText("Could not run " + filename + ":\n\n" + msg)
            m.setIcon(m.Icon.Warning)
            m.exec()

    def _runNotebook(self):
        filename = self._item.path()
        shell = pyzo.shells.getCurrentShell()
        if shell is not None:
            shell.restart(filename)
        else:
            msg = "No shell to run notebook in. "
            m = QtWidgets.QMessageBox(self)
            m.setWindowTitle(translate("menu dialog", "Could not run notebook"))
            m.setText("Could not run " + filename + ":\n\n" + msg)
            m.setIcon(m.Icon.Warning)
            m.exec()

    def _importData(self):
        browser = self.parent().parent()
        wizard = browser.getImportWizard()
        wizard.open(self._item.path())

    def onDuplicate(self):
        return self._duplicateOrRename(False)

    def onRename(self):
        return self._duplicateOrRename(True)

    def onCreateFile(self):
        self._createDirOrFile(True)

    def onCreateDir(self):
        self._createDirOrFile(False)

    def _createDirOrFile(self, file=True):
        # Get title and label
        if file:
            title = translate("filebrowser", "Create new file")
            label = translate("filebrowser", "Give the new name for the file")
        else:
            title = translate("filebrowser", "Create new directory")
            label = translate("filebrowser", "Give the name for the new directory")

        # Ask for new filename
        name, ok = QtWidgets.QInputDialog.getText(
            self.parent(),
            title,
            label + ":\n{}".format(self._item.path()),
            QtWidgets.QLineEdit.EchoMode.Normal,
            "new name",
        )

        if name and ok:
            newpath = op.join(self._item.path(), name)
            if op.exists(newpath):
                QtWidgets.QMessageBox.warning(
                    self.parent(),
                    title,
                    translate(
                        "filebrowser",
                        "Operation aborted. A file with the same name already exists.",
                    ),
                    QtWidgets.QMessageBox.StandardButton.Ok,
                )
            else:
                # Push create task
                task = tasks.CreateTask(newpath=newpath, file=file)
                self._item._proxy.pushTask(task)

    def _duplicateOrRename(self, rename):
        # Get dirname and filename
        dirname, filename = op.split(self._item.path())

        # Get title and label
        title = translate("filebrowser", "Rename" if rename else "Duplicate")
        label = translate("filebrowser", "Give the new name for the file")
        if not rename:
            filename = "Copy of " + filename

        # Ask for new filename
        name, ok = QtWidgets.QInputDialog.getText(
            self.parent(),
            title,
            label + ":\n{}".format(self._item.path()),
            QtWidgets.QLineEdit.EchoMode.Normal,
            filename,
        )

        if name and ok:
            newpath = op.join(dirname, name)
            if op.exists(newpath):
                QtWidgets.QMessageBox.warning(
                    self.parent(),
                    title,
                    translate(
                        "filebrowser",
                        "Operation aborted. A file with the same name already exists.",
                    ),
                    QtWidgets.QMessageBox.StandardButton.Ok,
                )
            else:
                # Push rename resp. duplicate task
                task = tasks.RenameTask(newpath=newpath, removeold=rename)
                self._item._proxy.pushTask(task)

    def onDelete(self):
        SB = QtWidgets.QMessageBox.StandardButton
        b = QtWidgets.QMessageBox.question(
            self.parent(),
            translate("filebrowser", "Delete"),
            translate("filebrowser", "Are you sure that you want to delete")
            + ":\n{}".format(self._item.path()),
            SB.Yes | SB.Cancel,
        )
        # Push delete task
        if b == SB.Yes:
            self._item._proxy.pushTask(tasks.RemoveTask())



# -*- coding: utf-8 -*-
# This example demonstrates simple req rep.
#
# This time in event driven mode. This example only works locally, as
# we cannot start two event loops :)

## ========== One end

import yoton

verbosity = 0


# Create a replier class by subclassing RepChannel
class Reducer(yoton.RepChannel):
    def reduce(self, item):
        if item == 2:
            raise ValueError("I do not like 2.")
        return item - 1


# Create a context and a rep channel
ct1 = yoton.Context(verbose=verbosity)
rep = Reducer(ct1, "reduce")

# Connect and turn duplicator on
ct1.bind("publichost:test")
rep.set_mode("event")


## ========== Other end

import yoton
import time

verbosity = 0

# Create a context and a req channel
ct2 = yoton.Context(verbose=verbosity)
req = yoton.ReqChannel(ct2, "reduce")

# Connect
ct2.connect("publichost:test")


# Create reply handler and bind it
def reply_handler(future):
    # Check error, cancelled, or get number
    if future.exception():
        # Calling result() would raise the exception, so let's just
        # print it and make up our own number
        print("oops: " + str(future.exception()))
        number = 1
    elif future.cancelled():
        print("oops: request was cancelled.")
    else:
        number = future.result()

    if number > 0:
        print("we now have %i." % number)
        time.sleep(0.5)
        new_future = req.reduce(number)
        new_future.add_done_callback(reply_handler)
    else:
        print("Done")
        yoton.stop_event_loop()


# Send first message
new_future = req.reduce(7)
new_future.add_done_callback(reply_handler)

# Enter event loop
yoton.start_event_loop()



"""yoton.clientserver.py

Yoton comes with a small framework to setup a request-reply pattern
using a client-server model (over a non-persistent connection),
similar to telnet. This allows one process to easily ask small pieces
of information from another process.

To create a server, create a class that inherits from
yoton.RequestServer and implement its handle_request() method.

A client process can simply use the yoton.do_request function.
Example: ``yoton.do_request('www.google.com:80', 'GET http/1.1\\r\\n')``

The client server model is implemented using one function and one class:
yoton.do_request and yoton.RequestServer.

Details
-------

The server implements a request/reply pattern by listening at a socket.
Similar to telnet, each request is handled using a connection
and the socket is closed after the response is sent.

The request server can setup to run in the main thread, or can be started
using its own thread. In the latter case, one can easily create multiple
servers in a single process, that listen on different ports.

"""

import time
import socket
import threading

from yoton.misc import basestring, str
from yoton.misc import split_address, getErrorMsg
from yoton.core import send_all, recv_all


class RequestServer(threading.Thread):
    """RequestServer(address, async_val=False, verbose=0)

    Setup a simple server that handles requests similar to a telnet server,
    or asyncore. Starting the server using run() will run the server in
    the calling thread. Starting the server using start() will run the
    server in a separate thread.

    To create a server, subclass this class and re-implement the
    handle_request method. It accepts a request and should return a
    reply. This server assumes utf-8 encoded messages.

    Parameters
    ----------
    address : str
        Should be of the shape hostname:port.
    async_val : bool
        If True, handles each incoming connection in a separate thread.
        This might be advantageous if a the handle_request() method
        takes a long time to execute.
    verbose : bool
        If True, print a message each time a connection is accepted.

    Notes on hostname
    -----------------
    The hostname can be:
      * The IP address, or the string hostname of this computer.
      * 'localhost': the connection is only visible from this computer.
        Also some low level networking layers are bypassed, which results
        in a faster connection. The other context should also connect to
        'localhost'.
      * 'publichost': the connection is visible by other computers on the
        same network.

    """

    def __init__(self, address, async_val=False, verbose=0):
        threading.Thread.__init__(self)

        # Store whether to handle requests asynchronously
        self._async = async_val

        # Verbosity
        self._verbose = verbose

        # Determine host and port (assume tcp)
        protocol, host, port = split_address(address)

        # Create socket. Apply SO_REUSEADDR when binding, so that a
        # improperly closed socket on the same port will not prevent
        # us connecting.
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Bind (can raise error if port is not available)
        try:
            s.bind((host, port))
        except Exception:
            s.close()
            raise

        # Store socket instance
        self._bsd_socket = s

        # To stop serving
        self._stop_me = False

        # Make daemon
        self.daemon = True

    def start(self):
        """start()
        Start the server in a separate thread.
        """
        self._stop_me = False
        threading.Thread.start(self)

    def stop(self):
        """stop()
        Stop the server.
        """
        self._stop_me = True

    def run(self):
        """run()
        The server's main loop.
        """

        # Get socket instance
        s = self._bsd_socket

        # Notify
        hostname, port = s.getsockname()
        if self._verbose:
            print("Yoton: hosting at %s, port %i" % (hostname, port))

        # Tell the socket it is a host, accept multiple
        s.listen(1)

        # Set timeout so that we can check _stop_me from time to time
        self._bsd_socket.settimeout(0.25)

        # Enter main loop
        while not self._stop_me:
            try:
                s, addr = self._bsd_socket.accept()
            except socket.timeout:
                pass
            except InterruptedError:
                pass
            else:
                # Show handling?
                if self._verbose:
                    print("handling request from: " + str(addr))
                # Handle request
                if self._async:
                    rh = SocketHandler(self, s)
                    rh.start()
                else:
                    self._handle_connection(s)

        # Close down
        try:
            self._bsd_socket.close()
        except socket.error:
            pass

    def _handle_connection(self, s):
        """_handle_connection(s)
        Handle an incoming connection.
        """
        try:
            self._really_handle_connection(s)
        except Exception:
            print("Error handling request:")
            print(getErrorMsg())

    def _really_handle_connection(self, s):
        """_really_handle_connection(s)
        Really handle an incoming connection.
        """
        # Get request
        request = recv_all(s, True)
        if request is None:
            return

        # Get reply
        reply = self.handle_request(request)

        # Test
        if not isinstance(reply, basestring):
            raise ValueError("handle_request() should return a string.")

        # Send reply
        send_all(s, reply, True)

        # Close the socket
        try:
            s.close()
        except socket.error:
            pass

    def handle_request(self, request):
        """handle_request(request)

        Return a reply, given the request. Overload this method to create
        a server.

        The standard implementation echos the request, waits one second
        when receiving 'wait' and stops the server when receiving 'stop'.

        """
        # Special cases
        if request == "wait":
            time.sleep(1.0)
        elif request == "stop":
            self._stop_me = True

        # Echo
        return "Requested: " + request


class SocketHandler(threading.Thread):
    """SocketHandler(server, s)
    Simple thread that handles a connection.
    """

    def __init__(self, server, s):
        threading.Thread.__init__(self)
        self._server = server
        self._bsd_socket = s

    def run(self):
        self._server._handle_connection(self._bsd_socket)


def do_request(address, request, timeout=-1):
    """do_request(address, request, timeout=-1)

    Do a request at the server at the specified address. The server can
    be a yoton.RequestServer, or any other server listening on a socket
    and following a REQ/REP pattern, such as html or telnet. For example:
    ``html = do_request('www.google.com:80', 'GET http/1.1\\r\\n')``

    Parameters
    ----------
    address : str
        Should be of the shape hostname:port.
    request : string
        The request to make.
    timeout : float
        If larger than 0, will wait that many seconds for the response, and
        return None if timed out.

    Notes on hostname
    -----------------
    The hostname can be:
      * The IP address, or the string hostname of this computer.
      * 'localhost': the connection is only visible from this computer.
        Also some low level networking layers are bypassed, which results
        in a faster connection. The other context should also connect to
        'localhost'.
      * 'publichost': the connection is visible by other computers on the
        same network.

    """

    # Determine host (assume tcp)
    protocol, host, port = split_address(address)

    # Check request
    if not isinstance(request, basestring):
        raise ValueError("request should be a string.")

    # Check timeout
    if timeout is None:
        timeout = -1

    # Create socket and connect
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
    except socket.error:
        s.close()
        raise RuntimeError("No server is listening at the given port.")

    # Send request
    send_all(s, request, True)

    # Receive reply
    reply = recv_all(s, timeout)

    # Close socket
    try:
        s.close()
    except socket.error:
        pass

    # Done
    return reply


if __name__ == "__main__":

    class Lala(RequestServer):
        def handle_request(self, req):
            print("REQ:", repr(req))
            return "The current time is %i" % time.time()

    s = Lala("localhost:test", 0, 1)
    s.start()
    if False:
        print(do_request("localhost:test", "wait", 5))
        for i in range(10):
            print(do_request("localhost:test", "hi" + str(i)))
    # do_request('localhost:test', 'wait'); do_request('localhost:test', 'hi');



import re
import sys
import time
import subprocess


r"""
Some explanations regarding the progress bars
=============================================

When executing "pip install ..." or "pip download ..." pip will display a progress bar
for the download progress in the terminal.

For older pip versions it was enough to just fordward the stdout of pip to the Pyzo shell
to have a continuosly updating progress bar.
But this stopped working with newer pip versions -- the progressbar was only displayed
once when the whole download was already completed.


With the curernt pip release v25.2 we have three ways to display the progress bar:
1) use no workarounds, as in Pyzo <= v4.20:
    The progressbar will only be displayed when finished.
    --> bad user experience, because pip seems to be stuck for larger downloads
2) use advanced terminal escape sequences
    We can do this by setting some environment variables when calling pip via subprocess:
        TERM="" (instead of "dumb")
        TTY_COMPATIBLE=1
    We would also have to implement or cleverly replace escape sequences such as:
        b'\x1b[?25l'
        b'\x1b[2K'
        b'\x1b[?25h'
    This would require some bigger workarounds, and future pip releases could make use of
    different escape sequences.
    --> not a stable interface
3) use raw progress bars, which were introduced in pip v24.1 (2024-06-20)
    see https://github.com/pypa/pip/issues/11508
    To use these, we need to pass "--progress-bar=raw" to pip, if pip is new enough.
    Instead of the rendered progress bar, pip will write simple lines such as
        b'Progress 19398656 of 73160002\n'
    for each update.
    In Pyzo, we parse these lines and render our own progress bar.

--> we use 3)


more information about pip:
    https://github.com/pypa/pip
    https://pip.pypa.io/en/stable/news/
"""


def datasize_to_string(size_bytes):
    size_bytes = float(size_bytes)  # only needed for Python 2.7
    prefixes = ["", "k", "M", "G", "T"]
    k = 1000
    precision = 0 if size_bytes < k else 1
    for i, prefix in enumerate(prefixes):
        v = size_bytes / k**i
        if v < k:
            break
    return "{:.{}f} {}B".format(v, precision, prefix)


def build_progress_line(current, total, fill_char):
    bar_length = 40
    if total > 0 and current >= 0:
        percent = 100 * min(1, max(0, current / total))
    else:
        percent = 0
    filled_length = round(percent / 100 * bar_length)
    bar = fill_char * filled_length + " " * (bar_length - filled_length)
    line = "[{}] {:3d} %   {} of {}".format(
        bar, round(percent), datasize_to_string(current), datasize_to_string(total)
    )
    return line


def subprocess_with_callback(callback, cmd, use_raw_progress_bar, **kwargs):
    """Execute command in subprocess, stdout is passed to the
    callback function. Returns the returncode of the process.
    If callback is None, simply prints any stdout.
    """

    # Set callback to void if None
    if callback is None:
        callback = lambda x: None

    # Execute command
    try:
        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs
        )
    except OSError:
        type, err, tb = sys.exc_info()
        del tb
        callback(str(err) + "\n")
        return -1

    progress_bar_dash = b"\xe2\x94\x81"  # U+2501, a thick horizontal line character
    pending = b""
    last_line_was_raw_progress_bar = False
    while p.poll() is None:
        time.sleep(0.001)
        # Read text and process
        c = p.stdout.read(1)
        pending += c

        if use_raw_progress_bar:
            if c == b"\n":
                # instead of
                # mo = re.fullmatch(rb"Progress (\d+) of (\d+)\r?\n", pending)
                # we use
                mo = re.match(b"^Progress (\\d+) of (\\d+)\\r?\\n$", pending)
                # because of Python 2.7

                if mo:
                    current = int(mo[1])
                    total = int(mo[2])
                    pending = build_progress_line(
                        current, total, progress_bar_dash.decode("utf-8")
                    ).encode("utf-8")
                    if last_line_was_raw_progress_bar:
                        pending = b"\r" + pending
                    last_line_was_raw_progress_bar = True
                else:
                    if last_line_was_raw_progress_bar:
                        pending = b"\n" + pending
                    last_line_was_raw_progress_bar = False

                callback(pending.decode("utf-8", "ignore"))
                pending = b""
        else:
            if c in b"-.\n" or pending.endswith(progress_bar_dash):
                callback(pending.decode("utf-8", "ignore"))
                pending = b""

    # Process remaining text
    pending += p.stdout.read()
    if last_line_was_raw_progress_bar:
        pending = b"\n" + pending

    callback(pending.decode("utf-8", "ignore"))
    p.stdout.close()  # avoid ResourceWarning: unclosed file <_io.BufferedReader>

    # Done
    return p.returncode


def print_(p):
    sys.stdout.write(p)
    sys.stdout.flush()


def pip_command(*args):
    """Do a pip command, e.g. "install networkx".
    Installs in the current interpreter.
    """
    args = list(args)
    use_raw_progress_bar = True
    if len(args) > 0 and args[0] in ["install", "download"]:
        if get_pip_version() >= (24, 1):
            args.insert(1, "--progress-bar=raw")
        else:
            use_raw_progress_bar = False

    # By using use_raw_progress_bar=True as a default,
    # writing to the Pyzo shell will only be performed on '\n'.
    # --> better performance for "pip list" etc.

    cmd = [sys.executable, "-m", "pip"] + args
    subprocess_with_callback(print_, cmd, use_raw_progress_bar)


def get_pip_version():
    """Returns a tuple of mayor and minor version of the pip module."""
    version = (0, 0)  # fallback value
    cmd = [sys.executable, "-m", "pip", "--version"]
    try:
        returned_bytes = subprocess.check_output(cmd)  # e.g. b'pip 25.2 from ...'
        version_string = returned_bytes.split(None, 2)[1].decode("ascii")
        version = tuple(int(s) for s in version_string.split(".")[:2])
    except Exception:
        pass
    return version


if __name__ == "__main__":
    pip_command("install", "networkx")



from pyzo import translate

tool_name = translate("pyzoFileBrowser", "File Browser")
tool_summary = "Browse the files in your projects."


""" File browser tool

A powerfull tool for managing your projects in a lightweight manner.
It has a few file management utilities as well.


Config
======

The config consists of three fields:

  * list expandedDirs, with each element a directory
  * list starredDirs, with each element a dict with fields:
      * str path, the directory that is starred
      * str name, the name of the project (op.basename(path) by default)
      * bool addToPythonpath
  * searchMatchCase, searchRegExp, wholeWords, searchSubDirs, searchExcludeBinary
  * nameFilter

"""

# todo: List!
"""
  * see easily which files are opened (so it can be used as a secondary tab bar)
  * make visible the "current file" (if applicable)
  * single click on an file that is open selects it in the editor?
  * context menu items to run scripts
  * Support for multiple browsers.
"""

import sys
import os.path as op

import pyzo
from pyzo.util import zon as ssdf
from pyzo.qt import QtCore, QtGui, QtWidgets, API, QT_WRAPPER_VERSION_STR  # noqa

from .browser import Browser
from .utils import cleanpath, isdir


def _check_qt_version():
    if sys.platform == "darwin" and API == "PySide6":
        version = tuple(int(i) for i in QT_WRAPPER_VERSION_STR.split(".")[:3])
        if version == (6, 5, 1):
            print(
                "\nWARNING:On MacOS The file browser may be unstable with PySide6 version 6.5.1\n"
            )


class PyzoFileBrowser(QtWidgets.QWidget):
    """The main tool widget. An instance of this class contains one or
    more Browser instances. If there are more, they can be selected
    using a tab bar.
    """

    def __init__(self, parent):
        super().__init__(parent)

        _check_qt_version()

        # Get config
        toolId = self.__class__.__name__.lower() + "2"  # This is v2 of the file browser
        if toolId not in pyzo.config.tools:
            pyzo.config.tools[toolId] = ssdf.new()
        self.config = pyzo.config.tools[toolId]

        # Ensure three main attributes in config
        for name in ["expandedDirs", "starredDirs"]:
            if name not in self.config:
                self.config[name] = []

        # Ensure path in config
        if "path" not in self.config or not isdir(self.config.path):
            self.config.path = op.expanduser("~")

        # Check expandedDirs and starredDirs.
        # Make path objects and remove invalid dirs. Also normalize case,
        # should not be necessary, but maybe the config was manually edited.
        expandedDirs, starredDirs = [], []
        for d in self.config.starredDirs:
            if "path" in d and "name" in d and "addToPythonpath" in d:
                if isdir(d.path):
                    d.path = op.normcase(cleanpath(d.path))
                    starredDirs.append(d)
        for p in set([str(p) for p in self.config.expandedDirs]):
            if isdir(p):
                p = op.normcase(cleanpath(p))
                # Add if it is a subdir of a starred dir
                for d in starredDirs:
                    if p.startswith(d.path):
                        expandedDirs.append(p)
                        break
        self.config.expandedDirs, self.config.starredDirs = expandedDirs, starredDirs

        # Create browser(s).
        self._browsers = [Browser(self, self.config)]

        # Layout
        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)
        layout.addWidget(self._browsers[0])
        layout.setSpacing(0)
        # set margins
        margin = pyzo.config.view.widgetMargin
        layout.setContentsMargins(margin, margin, margin, margin)

    def path(self):
        """Get the current path shown by the file browser."""
        browser = self._browsers[0]
        return browser._tree.path()

    def setPath(self, path):
        """Set the shown path."""
        browser = self._browsers[0]
        browser._tree.setPath(path)

    def setSearchText(self, needle, setFocus=False):
        """Set the text in the search field."""
        browser = self._browsers[0]
        browser.setSearchText(needle, setFocus)

    def getAddToPythonPath(self):
        """Returns the path to be added to the Python path when starting a shell
        If a project is selected, which has the addToPath checkbox selected,
        returns the path of the project. Otherwise, returns None
        """
        # Select browser
        browser = self._browsers[0]
        # Select active project
        d = browser.currentProject()
        if d and d.addToPythonpath:
            return d.path
        return None

    def getDefaultSavePath(self):
        """Returns the path to be used as default when saving a new file in pyzo.
        Or None if the no path could be determined
        """
        # Select current browser
        browser = self._browsers[0]
        # Select its path
        path = browser._tree.path()
        # Return
        if op.isabs(path) and isdir(path):
            return path

    def closeEvent(self, event):
        self.cleanUp()
        return super().closeEvent(event)

    def cleanUp(self):
        # Close all browsers so they can clean up the file system proxies
        for browser in self._browsers:
            browser.cleanUp()



"""Module yoton.channels.file

Defines a class that can be used to wrap a channel to give it
a file like interface.

"""

import sys
import os

from yoton.channels import PubChannel, SubChannel

PY2 = sys.version_info[0] == 2


class FileWrapper(object):
    """FileWrapper(channel, chunksize=0, echo=None)

    Class that wraps a PubChannel or SubChannel instance to provide
    a file-like interface by implementing methods such as read() and
    write(), and other stuff specified in:
    [[http://docs.python.org/library/stdtypes.html#bltin-file-objects]]

    The file wrapper also splits messages into smaller messages if they
    are above the chunksize (only if chunksize > 0).

    On Python 2, the read methods return str (utf-8 encoded Unicode).

    """

    # Our file-like objects should not implement:
    # explicitly stated: fileno, isatty
    # don't seem to make sense: readlines, seek, tell, truncate, errors,
    # mode, name,

    def __init__(self, channel, chunksize=0, echo=None, isatty=False):
        if not isinstance(channel, (PubChannel, SubChannel)):
            raise ValueError("FileWrapper needs a PubChannel or SubChannel.")
        if echo is not None:
            if not isinstance(echo, PubChannel):
                raise ValueError("FileWrapper echo needs to be a PubChannel.")

        self._channel = channel
        self._chunksize = int(chunksize)
        self._echo = echo
        self._pid = os.getpid()  # To detect whether we are in multi-process
        self.errors = "strict"  # compat
        self._isatty = isatty

    def close(self):
        """Close the file object."""
        # Deal with multiprocessing
        if self._pid != os.getpid():
            if self is sys.stdin:
                sys.__stdin__.close()
            elif self is sys.stdout:
                sys.__stdout__.close()
            elif self is sys.stderr:
                sys.__stderr__.close()
            return
        # Normal behavior
        self._channel.close()

    @property
    def encoding(self):
        """The encoding used to encode strings to bytes and vice versa."""
        return "UTF-8"

    @property
    def closed(self):
        """Get whether the file is closed."""
        return self._channel._closed

    def flush(self):
        """Wait here until all messages have been sent."""
        self._channel._context.flush()

    @property
    def newlines(self):
        """The type of newlines used. Returns None; we never know what the
        other end could be sending!
        """
        return None

    # this is for the print statement to keep track spacing stuff
    @property
    def softspace(self):
        return getattr(self, "_softspace", False)

    @softspace.setter
    def softspace(self, value):
        self._softspace = bool(value)

    def read(self, block=None):
        """Alias for recv()."""
        res = self._channel.recv(block)
        if res and self._echo is not None:
            self._echo.send(res)
        if PY2:
            return res.encode("utf-8")
        else:
            return res

    def write(self, message):
        """Uses channel.send() to send the message over the Yoton network.
        The message is partitioned in smaller parts if it is larger than
        the chunksize.
        """
        # Deal with multiprocessing
        if self._pid != os.getpid():
            realfile = None
            if self is sys.stdout:
                realfile = sys.__stdout__
            elif self is sys.stderr:
                realfile = sys.__stderr__
            if realfile is not None:
                realfile.write(message)
                realfile.flush()
            return

        chunkSize = self._chunksize
        if 0 < chunkSize < len(message):
            for i in range(0, len(message), chunkSize):
                self._channel.send(message[i : i + chunkSize])
        else:
            self._channel.send(message)

    def writelines(self, lines):
        """Write a sequence of messages to the channel."""
        for line in lines:
            self._channel.send(line)

    def readline(self, size=0):
        """Read one string that was sent as one from the other end (always
        in blocking mode). A newline character is appended if it does not
        end with one.

        If size is given, returns only up to that many characters, the rest
        of the message is thrown away.
        """

        # Get line
        line = self._channel.recv(True)

        # Echo
        if line and self._echo is not None:
            self._echo.send(line)

        # Make sure it ends with newline
        if not line.endswith("\n"):
            line += "\n"

        # Decrease size?
        if size:
            line = line[:size]

        # Done
        if PY2:
            return line.encode("utf-8")
        else:
            return line

    def isatty(self):
        """Get whether this is a terminal."""
        return self._isatty



#!/usr/bin/env python3

"""This script runs a test for the code editor component."""

import os
import sys
from qt import QtGui, QtCore, QtWidgets

Qt = QtCore.Qt

## Go up one directory and then import the codeeditor package

os.chdir("..")
sys.path.insert(0, ".")
from codeeditor import CodeEditor


if __name__ == "__main__":
    app = QtWidgets.QApplication([])

    # Create editor instance
    e = CodeEditor(
        highlightCurrentLine=True,
        longLineIndicatorPosition=20,
        showIndentationGuides=True,
        showWhitespace=True,
        showLineEndings=True,
        wrap=True,
        showLineNumbers=True,
    )

    QtWidgets.QShortcut(QtGui.QKeySequence("F1"), e).activated.connect(
        e.autocompleteShow
    )
    QtWidgets.QShortcut(QtGui.QKeySequence("F2"), e).activated.connect(
        e.autocompleteCancel
    )
    QtWidgets.QShortcut(QtGui.QKeySequence("F3"), e).activated.connect(
        lambda: e.calltipShow(0, "test(foo, bar)")
    )
    QtWidgets.QShortcut(QtGui.QKeySequence("Shift+Tab"), e).activated.connect(
        e.dedentSelection
    )  # Shift + Tab

    # TODO: somehow these shortcuts don't work in this test-app, but they do in
    # pyzo. May have something to do with overriding slots of Qt-native objects?
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+C"), e).activated.connect(e.copy)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+X"), e).activated.connect(e.cut)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+V"), e).activated.connect(e.paste)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Shift+V"), e).activated.connect(
        e.pasteAndSelect
    )
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Z"), e).activated.connect(e.undo)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Y"), e).activated.connect(e.redo)

    e.setPlainText(
        "foo(bar)\nfor bar in range(5):\n  print bar\n"
        + "\nclass aap:\n  def monkey(self):\n    pass\n\n"
    )

    # Run application
    e.show()
    s = QtWidgets.QSplitter()
    s.addWidget(e)
    s.addWidget(QtWidgets.QLabel("test"))
    s.show()
    app.exec()



"""Module main

This module contains the main frame. Implements the main window.
Also adds some variables to the pyzo namespace, such as the callLater
function which is also defined here.

"""

import os
import sys
import time
import base64
from queue import Queue, Empty

import pyzo
from pyzo.core.icons import IconArtist
from pyzo.core import commandline
from pyzo.core.statusbar import StatusBar
from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo.core.splash import SplashWidget
from pyzo.util import paths
from pyzo.util import zon as ssdf  # zon is ssdf-light
from pyzo import translate
from pyzo.core.views import ViewManager


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, parent=None, locale=None):
        super().__init__(parent)

        self._closeflag = 0  # Used during closing/restarting

        # Init window title and application icon
        # Set title to something nice. On Ubuntu 12.10 this text is what
        # is being shown at the fancy title bar (since it's not properly
        # updated)
        self.setMainTitle()
        loadAppIcons()
        self.setWindowIcon(pyzo.icon)

        # Restore window geometry before drawing for the first time,
        # such that the window is in the right place
        self.resize(800, 600)  # default size
        self.restoreGeometry()

        # Show splash screen (we need to set our color too)
        w = SplashWidget(self, distro="no distro")
        self.setCentralWidget(w)
        self.setStyleSheet("QMainWindow { background-color: #268bd2;}")

        # Show empty window and disable updates for a while
        self.show()
        self.paintNow()
        self.setUpdatesEnabled(False)

        # Determine timeout for showing splash screen
        splash_timeout = time.time() + 1.0

        # Set locale of main widget, so that qt strings are translated
        # in the right way
        if locale:
            self.setLocale(locale)

        # Store myself
        pyzo.main = self

        # Init dockwidget settings
        self.setTabPosition(
            QtCore.Qt.DockWidgetArea.AllDockWidgetAreas,
            QtWidgets.QTabWidget.TabPosition.South,
        )
        self.setDockOptions(
            QtWidgets.QMainWindow.DockOption.AllowTabbedDocks
            | QtWidgets.QMainWindow.DockOption.AllowNestedDocks
            # |  QtWidgets.QMainWindow.DockOption.AnimatedDocks
        )

        # Set window atrributes
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_AlwaysShowToolTips, True)

        # Load icons and fonts
        loadIcons()
        loadFonts()

        # Set qt style, test success and detect Qt dark mode
        self.setQtStyle(None)  # None means init!

        if pyzo.config.settings.theme == "":
            # set light or dark mode when running Pyzo the first time
            pyzo.config.settings.theme = "default_dark" if pyzo.darkQt else "default"

        # detect dark syntax style (light text on dark background)
        try:
            theme = pyzo.themes[pyzo.config.settings.theme.lower()]["data"]
            s = theme["editor.text"]  # e.g.: "fore:#657b83, back:#fff"
            colors = dict(
                [tuple(s3.strip() for s3 in s2.split(":")) for s2 in s.split(",")]
            )
            pyzo.darkSyntax = (
                QtGui.QColor(colors["fore"]).lightness()
                > QtGui.QColor(colors["back"]).lightness()
            )
        except KeyError:
            pyzo.darkSyntax = False

        # Hold the splash screen if needed
        while time.time() < splash_timeout:
            QtWidgets.qApp.sendPostedEvents()
            QtWidgets.qApp.processEvents()
            time.sleep(0.05)

        # Populate the window (imports more code)
        self._populate()

        # Revert to normal background, and enable updates
        self.setStyleSheet("")
        self.setUpdatesEnabled(True)

        # Restore window state, force updating, and restore again
        self.restoreState()
        self.paintNow()
        self.restoreState()
        pyzo.editors.restoreEditorState()

        # Instantiate view manager
        pyzo.viewManager = ViewManager()

        # Present user with wizard if he/she is new.
        if False:  # pyzo.config.state.newUser:
            from pyzo.util.pyzowizard import PyzoWizard

            w = PyzoWizard(self)
            w.show()  # Use show() instead of exec() so the user can interact with pyzo

        # Create new shell config if there is None
        if not pyzo.config.shellConfigs2:
            from pyzo.core.kernelbroker import KernelInfo

            pyzo.config.shellConfigs2.append(KernelInfo())

        # Focus on editor
        e = pyzo.editors.getCurrentEditor()
        if e is not None:
            e.setFocus()

        # Handle any actions
        commandline.handle_cmd_args()

    # To force drawing ourselves
    def paintEvent(self, event):
        super().paintEvent(event)
        self._ispainted = True

    def paintNow(self):
        """Enforce a repaint and keep calling processEvents until
        we are repainted.
        """
        self._ispainted = False
        self.update()
        while not self._ispainted:
            QtWidgets.qApp.sendPostedEvents()
            QtWidgets.qApp.processEvents()
            time.sleep(0.01)

    def _populate(self):
        # Delayed imports
        from pyzo.core.editorTabs import EditorTabs
        from pyzo.core.shellStack import ShellStackWidget
        from pyzo.core import codeparser
        from pyzo.core.history import CommandHistory
        from pyzo.tools import ToolManager

        # Instantiate tool manager
        pyzo.toolManager = ToolManager()

        # Instantiate and start source-code parser
        if pyzo.parser is None:
            pyzo.parser = codeparser.Parser()
            pyzo.parser.start()

        # Create editor stack and make the central widget
        pyzo.editors = EditorTabs(self)
        self.setCentralWidget(pyzo.editors)

        # Create floater for shell
        self._shellDock = dock = QtWidgets.QDockWidget(self)
        DWF = dock.DockWidgetFeature
        if pyzo.config.settings.allowFloatingShell:
            dock.setFeatures(DWF.DockWidgetMovable | DWF.DockWidgetFloatable)
        else:
            dock.setFeatures(DWF.DockWidgetMovable)
        dock.setObjectName("shells")
        dock.setWindowTitle("Shells")
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, dock)

        # Create shell stack
        pyzo.shells = ShellStackWidget(self)
        dock.setWidget(pyzo.shells)

        # Initialize command history
        pyzo.command_history = CommandHistory("command_history.py")

        # Create the default shell when returning to the event queue
        callLater(pyzo.shells.addShell)

        # Create status bar
        self.setStatusBar(StatusBar())
        # show or hide
        self.statusBar().setVisible(pyzo.config.view.showStatusbar)

        # Create menu
        from pyzo.core import menu

        pyzo.keyMapper = menu.KeyMapper()
        menu.buildMenus(self.menuBar())

        # Add the context menu to the editor
        pyzo.editors.addContextMenu()
        pyzo.shells.addContextMenu()

        # Load tools
        if pyzo.config.state.newUser and not pyzo.config.state.loadedTools:
            pyzo.toolManager.loadTool("pyzosourcestructure")
            pyzo.toolManager.loadTool("pyzofilebrowser", "pyzosourcestructure")
        elif pyzo.config.state.loadedTools:
            self.restoreTools()

    def setMainTitle(self, path=None):
        """Set the title of the main window, by giving a file path."""
        if not path:
            # Plain title
            title = "Interactive Editor for Python"
        else:
            # Title with a filename
            name = os.path.basename(path)
            if os.path.isfile(path):
                pass
            elif name == path:
                path = translate("main", "unsaved")
            else:
                pass  # We hope the given path is informative
            # Set title
            tmp = {
                "fileName": name,
                "filename": name,
                "name": name,
                "fullPath": path,
                "fullpath": path,
                "path": path,
            }
            title = pyzo.config.advanced.titleText.format(**tmp)

        # Set
        self.setWindowTitle(title)

    def getWindowState(self):
        return {
            # list of tools which are loaded
            "loadedTools": pyzo.toolManager.getLoadedTools(),
            # geometry of the top level window
            "windowGeometry": base64.encodebytes(self.saveGeometry().data()).decode(
                "ascii"
            ),
            # layout of dockwidgets and toolbars
            "windowState": base64.encodebytes(self.saveState().data()).decode("ascii"),
        }

    def saveWindowState(self):
        """Save:
        * which tools are loaded
        * geometry of the top level windows
        * layout of dockwidgets and toolbars
        """
        pyzo.config.state.update(self.getWindowState())

    def restoreTools(self, value=None):
        """Restore loaded tools"""

        if value is None:
            # No value given, try to get it from the config
            value = pyzo.config.state.loadedTools

        toolsTarget = set(value)
        toolsLoaded = set(pyzo.toolManager.getLoadedTools())
        for t in toolsLoaded - toolsTarget:
            pyzo.toolManager.closeTool(t)
        for t in toolsTarget - toolsLoaded:
            pyzo.toolManager.loadTool(t)

    def restoreGeometry(self, value=None):
        """Restore window position and whether it is maximized"""

        if value is not None:
            return super().restoreGeometry(value)

        # No value given, try to get it from the config
        if pyzo.config.state.windowGeometry:
            try:
                geometry = pyzo.config.state.windowGeometry
                geometry = base64.decodebytes(geometry.encode("ascii"))
                self.restoreGeometry(geometry)
            except Exception as err:
                print("Could not restore window geometry: " + str(err))

    def restoreState(self, value=None):
        """Restore layout of dock widgets and toolbars"""

        if value is not None:
            return super().restoreState(value)

        # No value given, try to get it from the config
        if pyzo.config.state.windowState:
            try:
                state = pyzo.config.state.windowState
                state = base64.decodebytes(state.encode("ascii"))
                self.restoreState(state)
            except Exception as err:
                print("Could not restore window state: " + str(err))

    def setQtStyle(self, stylename=None):
        """Set the style and the palette, based on the given style name.
        If stylename is None or not given will do some initialization.
        If bool(stylename) evaluates to False will use the default style
        for this system. Returns the QStyle instance.
        """

        if stylename is None:
            # Initialize

            # Get native palette (used below)
            QtWidgets.qApp.nativePalette = QtWidgets.qApp.palette()

            # Obtain default style name
            pyzo.defaultQtStyleName = str(QtWidgets.qApp.style().objectName())

            # Other than gtk+ and mac, Fusion/Cleanlooks looks best (in my opinion)
            if "gtk" in pyzo.defaultQtStyleName.lower():
                pass  # Use default style
            elif "macintosh" in pyzo.defaultQtStyleName.lower():
                pass  # Use default style
            else:
                pyzo.defaultQtStyleName = "Fusion"

            # Set style if there is no style yet
            if not pyzo.config.view.qtstyle:
                pyzo.config.view.qtstyle = pyzo.defaultQtStyleName

        # Init
        if not stylename:
            stylename = pyzo.config.view.qtstyle

        # Check if this style exist, set to default otherwise
        styleNames = [name.lower() for name in QtWidgets.QStyleFactory.keys()]
        if stylename.lower() not in styleNames:
            stylename = pyzo.defaultQtStyleName

        # Try changing the style
        qstyle = QtWidgets.qApp.setStyle(stylename)

        # Set palette
        if qstyle:
            QtWidgets.qApp.setPalette(QtWidgets.qApp.nativePalette)

        # detect dark mode for widgets (might be different than dark mode for syntax)
        pal = QtWidgets.qApp.palette()
        pyzo.darkQt = (
            pal.window().color().lightness() < pal.windowText().color().lightness()
        )

        # Done
        return qstyle

    def closeEvent(self, event):
        """Override close event handler."""

        # Are we restaring?
        restarting = time.time() - self._closeflag < 1.0  # noqa: F841

        if pyzo.config.advanced.alwaysAskWhenClosingPyzo:
            dlg = QtWidgets.QMessageBox(pyzo.main)
            dlg.setWindowTitle("Pyzo")
            dlg.setText(translate("main", "Do you really want to close Pyzo?"))

            infoTexts = []
            busyStates = ("Busy", "Very busy", "Debug")
            if any(shell._state in busyStates for shell in pyzo.shells):
                infoTexts.append(
                    translate("main", "At least one shell is still executing code.")
                )
            if any(ed.document().isModified() for ed in pyzo.editors):
                infoTexts.append(
                    translate("main", "At least one editor has unsaved changes.")
                )
            dlg.setInformativeText("\n".join(infoTexts))

            btnClose = dlg.addButton(QtWidgets.QMessageBox.StandardButton.Close)
            btnCancel = dlg.addButton(QtWidgets.QMessageBox.StandardButton.Cancel)
            dlg.setDefaultButton(btnCancel)
            dlg.exec()
            if dlg.clickedButton() is not btnClose:
                self._closeflag = False
                event.ignore()
                return

        # Proceed with closing...
        result = pyzo.editors.closeAll()
        if not result:
            self._closeflag = False
            event.ignore()
            return
        else:
            self._closeflag = True
            # event.accept()  # Had to comment on Windows+py3.3 to prevent error

        # Save settings
        pyzo.saveConfig()
        pyzo.command_history.save()

        # Stop command server
        commandline.stop_our_server()

        # Proceed with closing shells
        pyzo.localKernelManager.terminateAll()

        for shell in pyzo.shells:
            shell._context.close()

        # The tools need to be explicitly closed to allow them to clean up
        for toolname in pyzo.toolManager.getLoadedTools():
            tool = pyzo.toolManager.getTool(toolname)
            if hasattr(tool, "cleanUp"):
                tool.cleanUp()

        # Stop all threads (this should really only be daemon threads)
        import threading

        for thread in threading.enumerate():
            if hasattr(thread, "stop"):
                try:
                    thread.stop(0.1)
                except Exception:
                    pass

        #         # Wait for threads to die ...
        #         # This should not be necessary, but I used it in the hope that it
        #         # would prevent the segfault on Python3.3. It didn't.
        #         timeout = time.time() + 0.5
        #         while threading.activeCount() > 1 and time.time() < timeout:
        #             time.sleep(0.1)
        #         print('Number of threads alive:', threading.activeCount())

        # Proceed as normal
        super().closeEvent(event)

    def restart(self):
        """Restart Pyzo."""

        self._closeflag = time.time()

        # Close
        self.close()

        if self._closeflag:
            # Get args
            args = list(sys.argv)

            if not paths.is_frozen():
                # Prepend the executable name (required on Linux)
                lastBit = os.path.basename(sys.executable)
                args.insert(0, lastBit)

            # When running from the pip entry point pyzo.exe ... (issue #641)
            if (
                len(args) == 2
                and args[0] == "python.exe"
                and not os.path.isfile(args[1])
            ):
                args = ["python.exe", "-m", "pyzo"]

            if sys.platform == "win32":
                # workaround for MSVCRT issue with spaces in arguments
                #     https://bugs.python.org/issue436259
                from subprocess import list2cmdline

                args = [list2cmdline([s]) for s in args]

            # Replace the process!
            os.execv(sys.executable, args)

    def createPopupMenu(self):
        # This menu pops up when right clicking on the menu bar
        # or when right clicking on the title bar of a panel.

        # We have to use the initialized menu from the parent class because creating
        # a completely new QMenu would result in a memory error.
        menu = super().createPopupMenu()
        menu.clear()

        # The build_callback function is necessary because without that, a simple lambda
        # would bind the argument to the latest action of the for loop via the closure.
        def build_callback(cb, action):
            return lambda: cb(action.isChecked())

        # Add all tools, with checkmarks for those that are active
        for tool in pyzo.toolManager.getToolInfo():
            a = menu.addAction(tool.name)
            a.setCheckable(True)
            a.setChecked(bool(tool.instance))
            a.triggered.connect(build_callback(tool.menuLauncher, a))

        return menu

    def keyPressEvent(self, event):
        # Here we forward the key press events to the editor widget so that tab switching
        # will also be possible when the focus is outside the editor widget.
        if pyzo.editors:  # check for None to avoid crash during Pyzo startup
            pyzo.editors.processKeyPressFromMainWindow(event)


def loadAppIcons():
    """Load the application icons."""
    # Get directory containing the icons
    appiconDir = os.path.join(pyzo.pyzoDir, "resources", "appicons")

    # Determine template for filename of the application icon-files.
    fnameT = "pyzologo{}.png"

    # Construct application icon. Include a range of resolutions. Note that
    # Qt somehow does not use the highest possible res on Linux/Gnome(?), even
    # the logo of qt-designer when alt-tabbing looks a bit ugly.
    pyzo.icon = QtGui.QIcon()

    # Construct another icon to show when the current shell is busy
    pyzo.iconRunning = QtGui.QIcon(pyzo.icon)

    for sze in reversed([16, 32, 48, 64, 128, 256]):
        fname = os.path.join(appiconDir, fnameT.format(sze))
        if os.path.isfile(fname):
            pyzo.icon.addFile(fname, QtCore.QSize(sze, sze))

            artist = IconArtist(pyzo.icon, size=sze)
            artist.setPenColor("#0D0")
            artist.setBrushColor("#0D0")
            a = 6 * sze // 16
            x2 = sze - 1
            x1 = x2 - a + 1
            y1 = sze - 1
            y2 = y1 - a + 1
            y3 = y2 - a + 1
            artist.addPolygon([(x1, y1), (x2, y2), (x1, y3)])
            pm = artist.finish().pixmap(sze, sze)
            pyzo.iconRunning.addPixmap(pm)

    # Set as application icon. This one is used as the default for all
    # windows of the application.
    QtWidgets.qApp.setWindowIcon(pyzo.icon)


def loadIcons():
    """Load all icons in the icon dir."""
    # Get directory containing the icons
    iconDir = os.path.join(pyzo.pyzoDir, "resources", "icons")

    # Construct other icons
    dummyIcon = IconArtist().finish()
    pyzo.icons = ssdf.new()
    for fname in os.listdir(iconDir):
        if fname.endswith(".png"):
            try:
                # Short and full name
                name = fname.split(".")[0]
                name = name.replace("pyzo_", "")  # discard prefix
                ffname = os.path.join(iconDir, fname)
                # Create icon
                icon = QtGui.QIcon()
                icon.addFile(ffname, QtCore.QSize(16, 16))
                # Store
                pyzo.icons[name] = icon
            except Exception as err:
                pyzo.icons[name] = dummyIcon
                print("Could not load icon {}: {}".format(fname, err))

    artist = IconArtist("folder_page")
    artist.addLayer("arrow_refresh")
    pyzo.icons["reload_file_from_disk"] = artist.finish()

    artist = IconArtist("page_white_copy")
    artist.addLayer("overlay_disk")
    pyzo.icons["save_copy_as"] = artist.finish()


def loadFonts():
    """Load all fonts that come with Pyzo."""
    import pyzo.codeeditor  # we need pyzo and codeeditor namespace here

    # Get directory containing the icons
    fontDir = os.path.join(pyzo.pyzoDir, "resources", "fonts")

    # Get database object
    db = QtGui.QFontDatabase  # static class

    # Set default font
    pyzo.codeeditor.Manager.setDefaultFontFamily("DejaVu Sans Mono")

    # Load fonts that are in the fonts directory
    if os.path.isdir(fontDir):
        for fname in os.listdir(fontDir):
            if "oblique" in fname.lower():  # issue #461
                continue
            if os.path.splitext(fname)[1].lower() in [".otf", ".ttf"]:
                try:
                    db.addApplicationFont(os.path.join(fontDir, fname))
                except Exception as err:
                    print("Could not load font {}: {}".format(fname, err))


class _CallbackEventHandler(QtCore.QObject):
    """Helper class to provide the callLater function."""

    def __init__(self):
        super().__init__()
        self.queue = Queue()

    def customEvent(self, event):
        while True:
            try:
                callback, args = self.queue.get_nowait()
            except Empty:
                break
            try:
                callback(*args)
            except Exception as why:
                print("callback failed: {}:\n{}".format(callback, why))

    def postEventWithCallback(self, callback, *args):
        self.queue.put((callback, args))
        QtWidgets.qApp.postEvent(self, QtCore.QEvent(QtCore.QEvent.Type.User))


def callLater(callback, *args):
    """Post a callback to be called in the main thread."""
    _callbackEventHandler.postEventWithCallback(callback, *args)


# Create callback event handler instance and insert function in pyzo namespace
_callbackEventHandler = _CallbackEventHandler()
pyzo.callLater = callLater


_SCREENSHOT_CODE = """
import random

numerator = 4

def get_number():
    # todo: something appears to be broken here
    val = random.choice(range(10))
    return numerator / val

class Groceries(list):
    \"\"\" Overloaded list class.
    \"\"\"
    def append_defaults(self):
        spam = 'yum'
        pie = 3.14159
        self.extend([spam, pie])

class GroceriesPlus(Groceries):
    \"\"\" Groceries with surprises!
    \"\"\"
    def append_random(self):
        value = get_number()
        self.append(value)

# Create some groceries
g = GroceriesPlus()
g.append_defaults()
g.append_random()

"""


def screenshotExample(width=1244, height=700):
    e = pyzo.editors.newFile()
    e.editor.setPlainText(_SCREENSHOT_CODE)
    pyzo.main.resize(width, height)


def screenshot(countdown=5):
    QtCore.QTimer.singleShot(countdown * 1000, _screenshot)


def _screenshot():
    # Grab
    print("SNAP!")
    screen = QtWidgets.qApp.primaryScreen()
    pix = screen.grabWindow(pyzo.main.winId())
    # Get name
    i = 1
    while i > 0:
        name = "pyzo_screen_{}_{:02d}.png".format(sys.platform, i)
        fname = os.path.join(os.path.expanduser("~"), name)
        if os.path.isfile(fname):
            i += 1
        else:
            i = -1
    # Save screenshot and a thumb
    pix.save(fname)
    thumb = pix.scaledToWidth(500, QtCore.Qt.SmoothTransformation)
    thumb.save(fname.replace("screen", "thumb"))
    print("Screenshot and thumb saved in", os.path.expanduser("~"))


pyzo.screenshot = screenshot
pyzo.screenshotExample = screenshotExample



import os
import sys
import time
import bdb
import traceback


# Patch bdb so that "stop debugging" stops code execution even when using
# "except Exception:" clauses, which catch the BdbQuit error.
# see https://github.com/python/cpython/issues/149309
if issubclass(bdb.BdbQuit, Exception):

    class BdbQuit(BaseException):
        pass

    bdb.BdbQuit = BdbQuit


class Debugger(bdb.Bdb):
    """Debugger for the pyzo kernel, based on bdb."""

    def __init__(self):
        self._wait_for_mainpyfile = False  # from pdb, do we need this?
        bdb.Bdb.__init__(self)
        self._debugmode = 0  # 0: no debug,  1: postmortem,  2: full debug
        self._files_with_offset = set()
        if hasattr(sys, "breakpointhook"):
            self._original_breakpointhook = sys.breakpointhook
            sys.breakpointhook = self.custom_breakpointhook
        self._last_db_command = None

    def custom_breakpointhook(self, *args, **kwargs):
        """handle "breakpoint()" commands
        see https://docs.python.org/3/library/sys.html#sys.breakpointhook
        """
        s = os.environ.get("PYTHONBREAKPOINT", None)
        if s == "0":
            pass  # deactivated
        elif s is None or s == "":
            self.set_trace()
        else:
            self._original_breakpointhook(*args, **kwargs)

    def clear_all_breaks(self):
        bdb.Bdb.clear_all_breaks(self)
        self._files_with_offset.clear()

    def trace_dispatch(self, frame, event, arg):
        # Overload to deal with offset in filenames
        # (cells or lines being executed)
        ori_filename = frame.f_code.co_filename

        if "+" in ori_filename and ori_filename not in self._files_with_offset:
            clean_filename, offset = ori_filename.rsplit("+", 1)
            try:
                offset = int(offset)
            except Exception:
                offset = None
            if offset is not None:
                # This is a cell or selected lines being executed
                self._files_with_offset.add(ori_filename)
                if clean_filename.startswith("<"):
                    self.fncache[ori_filename] = ori_filename
                for i in self.breaks.get(clean_filename, []):
                    self.set_break(ori_filename, i - offset)

        return bdb.Bdb.trace_dispatch(self, frame, event, arg)

    def interaction(self, frame, traceback_=None, pm=False, pm_exc=None):
        """Enter an interaction-loop for debugging. No GUI events are
        processed here. We leave this event loop at some point, after
        which the control flow will proceed.

        This is called to enter debug-mode at a breakpoint, or to enter
        post-mortem debugging.
        """
        interpreter = sys._pyzoInterpreter

        # Collect frames
        frames = []

        if pm:
            assert frame is None
            frame = traceback_.tb_frame

        # starting at the frame "frame", we add frames towards the callers:
        while frame:
            if frame is self.botframe:
                break
            co_filename = frame.f_code.co_filename
            if "pyzokernel" in co_filename:
                break  # pyzo kernel
            if "interactiveshell.py" in co_filename:
                break  # IPython kernel
            frames.append(frame)
            frame = frame.f_back
        frames.reverse()

        if pm:
            # For post-mortem, we also go towards the frame that caused the exception.
            # Some frames can have f_back set to None, like generator expressions
            # (see see https://stackoverflow.com/a/51867843), but we only go forward here.

            exc = pm_exc
            tb = traceback_
            while True:
                while tb:
                    frame = tb.tb_frame

                    # Same frames for different traceback entries can occur for example
                    # in nested try-except blocks. The traceback has entries for the same
                    # frame, but on different lines (e.g. one in the try-clause, one in the
                    # except clause).
                    if frames and frames[-1] is not frame:
                        frames.append(frame)
                    tb = tb.tb_next

                # "cause" vs "context" in exceptions that trigger other exceptions:
                #
                # try:
                #     something()
                # except Exception as e:
                #     raise TypeError('abc') from e  # --> e was the cause
                #     # raise TypeError('abc')  # --> e was the context

                chained_exc = getattr(exc, "__cause__", None)

                # We could also check for context exceptions, but this might add more
                # stack frames than what is useful:
                #
                # if not chained_exc:
                #     chained_exc = getattr(exc, "__context__", None)

                if not chained_exc:
                    break
                # Older Python versions that have no exc.__traceback__ never reach this,
                # because they also have no __cause__ or __context__ attribute.
                exc = chained_exc
                tb = exc.__traceback__

        # Tell interpreter our stack
        if frames:
            interpreter._dbFrames = frames
            interpreter._dbFrameIndex = len(interpreter._dbFrames)
            frame = interpreter._dbFrames[interpreter._dbFrameIndex - 1]
            interpreter._dbFrameName = frame.f_code.co_name
            interpreter.locals = frame.f_locals
            interpreter.globals = frame.f_globals

        # Let the IDE know
        self._debugmode = 1 if pm else 2
        self.writestatus()

        # Enter interact loop. We may hang in here for a while ...
        self._interacting = True
        while self._interacting:
            try:
                time.sleep(0.05)
                interpreter.process_commands()
                pe = os.getenv("PYZO_PROCESS_EVENTS_WHILE_DEBUGGING", "").lower()
                if pe in ("1", "true", "yes"):
                    interpreter.guiApp.process_events()
            except KeyboardInterrupt:
                self.message("KeyboardInterrupt")

        # Reset
        self._debugmode = 0
        interpreter.locals = interpreter._main_locals
        interpreter.globals = interpreter._main_globals
        interpreter._dbFrames = []
        self.writestatus()

    def stopinteraction(self):
        """Stop the interaction loop."""
        self._interacting = False
        sys.settrace(None)

    def set_on(self):
        """To turn debugging on right before executing code."""
        # Reset and set bottom frame
        self.reset()
        self.botframe = sys._getframe().f_back
        # Don't stop except at breakpoints or when finished
        self._set_stopinfo(self.botframe, None, -1)
        self._last_db_command = None

    def message(self, msg):
        """Alias for interpreter.write(), but appends a newline.
        Writes to stderr.
        """
        sys._pyzoInterpreter.write(msg + "\n")

    def error(self, msg):
        """method used in some code that we copied from pdb."""
        raise self.message("*** " + msg)

    def writestatus(self):
        """Write the debug status so the IDE can take action."""

        interpreter = sys._pyzoInterpreter

        # Collect frames info
        frames = []
        for f in interpreter._dbFrames:
            # Get fname and lineno, and correct if required
            fname, lineno = f.f_code.co_filename, f.f_lineno
            fname, lineno = interpreter.correctfilenameandlineno(fname, lineno)
            if not fname.startswith("<"):
                fname2 = os.path.abspath(fname)
                if os.path.isfile(fname2):
                    fname = fname2
            frames.append((fname, lineno, f.f_code.co_name))
            # Build string
            # text = 'File "%s", line %i, in %s' % (
            #                        fname, lineno, f.f_code.co_name)
            # frames.append(text)

        # Send info object
        state = {
            "index": interpreter._dbFrameIndex,
            "frames": frames,
            "debugmode": self._debugmode,
        }
        interpreter.context._stat_debug.send(state)

    ## Stuff that we need to overload

    # Overload set_break to also allow non-existing filenames like "<tmp 1"
    def set_break(self, filename, lineno, temporary=False, cond=None, funcname=None):
        filename = self.canonic(filename)
        list = self.breaks.setdefault(filename, [])
        if lineno not in list:
            list.append(lineno)
        bdb.Breakpoint(filename, lineno, temporary, cond, funcname)

    # Prevent stopping in bdb code or pyzokernel code
    def stop_here(self, frame):
        result = bdb.Bdb.stop_here(self, frame)
        if result:
            return ("bdb.py" not in frame.f_code.co_filename) and (
                "pyzokernel" not in frame.f_code.co_filename
            )

    def do_clear(self, arg):
        """ """  # this docstring is a single space on purpose
        # Clear breakpoints, we need to overload from Bdb,
        # but do not expose this command to the user.
        """cl(ear) filename:lineno\ncl(ear) [bpnumber [bpnumber...]]
        With a space separated list of breakpoint numbers, clear
        those breakpoints.  Without argument, clear all breaks (but
        first ask confirmation).  With a filename:lineno argument,
        clear all breaks at that line in that file.
        """
        if not arg:
            bplist = [bp for bp in bdb.Breakpoint.bpbynumber if bp]
            self.clear_all_breaks()
            for bp in bplist:
                self.message("Deleted %s" % bp)
            return
        if ":" in arg:
            # Make sure it works for "clear C:\foo\bar.py:12"
            i = arg.rfind(":")
            filename = arg[:i]
            arg = arg[i + 1 :]
            try:
                lineno = int(arg)
            except ValueError:
                err = "Invalid line number (%s)" % arg
            else:
                bplist = self.get_breaks(filename, lineno)
                err = self.clear_break(filename, lineno)
            if err:
                self.error(err)
            else:
                for bp in bplist:
                    self.message("Deleted %s" % bp)
            return
        numberlist = arg.split()
        for i in numberlist:
            try:
                bp = self.get_bpbynumber(i)
            except ValueError:
                self.error("Cannot get breakpoint by number.")
            else:
                self.clear_bpbynumber(i)
                self.message("Deleted %s" % bp)

    def user_call(self, frame, argument_list):
        """This method is called when there is the remote possibility
        that we ever need to stop in this function."""
        if self._wait_for_mainpyfile:
            return
        if self.stop_here(frame):
            self.message("--Call--")
            self.interaction(frame, None)

    def user_line(self, frame):
        """This function is called when we stop or break at this line."""
        if self._wait_for_mainpyfile:
            if (
                self.mainpyfile != self.canonic(frame.f_code.co_filename)
                or frame.f_lineno <= 0
            ):
                return
            self._wait_for_mainpyfile = False
        if True:  # self.bp_commands(frame):  from pdb
            self.interaction(frame, None)

    def user_return(self, frame, return_value):
        """This function is called when a return trap is set here."""
        if self._wait_for_mainpyfile:
            return
        frame.f_locals["__return__"] = return_value
        self.message("--Return--")
        self.interaction(frame, None)

    def user_exception(self, frame, exc_info):
        """This function is called if an exception occurs,
        but only if we are to stop at or just below this level."""
        if self._wait_for_mainpyfile:
            return
        exc_type, exc_value, exc_traceback = exc_info
        frame.f_locals["__exception__"] = exc_type, exc_value
        self.message(traceback.format_exception_only(exc_type, exc_value)[-1].strip())
        self.interaction(frame, exc_traceback)

    ## Commands

    def do_help(self, arg):
        """Get help on debug commands."""
        # Collect docstrings
        docs = {}
        for name in dir(self):
            if name.startswith("do_"):
                doc = getattr(self, name).__doc__
                if doc and doc != " ":
                    docs[name[3:]] = doc.strip()

        if not arg:
            print("All debug commands:")
            # Show docs in order
            for name in [
                "start",
                "stop",
                "frame",
                "up",
                "down",
                "next",
                "step",
                "return",
                "jump",
                "continue",
                "where",
                "events",
            ]:
                doc = docs.pop(name)
                name = name.rjust(10)
                print(" %s - %s" % (name, doc))
            # Show rest
            for name in docs:
                doc = docs[name]
                name = name.rjust(10)
                print(" %s - %s" % (name, doc))

        else:
            # Show specific doc
            name = arg.lower()
            doc = docs.get(name, None)
            if doc is not None:
                print("%s - %s" % (name, doc))
            else:
                print("Unknown debug command: %s" % name)

    def _get_exception_combinations(self, exc):
        """get all possible combinations for exceptions

        If the exception is an ExceptionGroup, it can consist of sub-exceptions
        and even nested ExceptionGroup objects. We identify a branch via the
        sub-exception numbers that are printed in the traceback ("---- 1 ----" etc.).

        A tuple (3, 2) addresses the ValueError sub-sub-exception in the following example.
        ExceptionGroup  --> ()
            ---- 1 ----  --> (1,)
            ZeroDivisionError
            ---- 2 ----  --> (2,)
            KeyError
            ---- 3 ----  --> (3,)
            ExceptionGroup
                ---- 1 ----  --> (3, 1)
                TypeError
                ---- 2 ----  --> (3, 2)
                ValueError

        return value: exc_combs, first_nongroup
            exc_combs ... a dictionary of kv-pairs
                key: the tuple branch-identifier
                value: the (sub-)exception
            first_nongroup ... the key to the exc_combs dict to the first exception
                that is not an ExceptionGroup -- useful as a default selection
        """
        exc_combs = {}
        first_nongroup = None

        if exc is not None:
            # add all combinations for nested exception groups
            p = ()
            stack = [(exc, p)]
            nongroup_exceptions = []
            exc_combs[p] = exc
            while len(stack) > 0:
                e, p = stack.pop()
                if e.__class__.__name__ == "ExceptionGroup":
                    for num, e2 in enumerate(e.exceptions, 1):
                        p2 = p + (num,)
                        stack.append((e2, p2))
                        exc_combs[p2] = e2
                else:
                    nongroup_exceptions.append(p)
            if nongroup_exceptions:
                first_nongroup = min(nongroup_exceptions)
        return exc_combs, first_nongroup

    def _get_selected_exception(self, exc, tb, arg):
        """get the selected exception of an ExceptionGroup"""
        exc_combs, first_nongroup = self._get_exception_combinations(exc)
        exc = exc_combs.get((), None)
        if len(exc_combs) > 1:
            # there is at least one ExceptionGroup
            user_selected_comb = None
            selected_comb = None
            if arg != "":
                try:
                    user_selected_comb = tuple([int(s) for s in arg.split()])
                    # Selecting the root-ExceptionGroup would normally be done with
                    # an empty tuple. But we want the empty tuple, which is args == "",
                    # to be the "automatic" selection option.
                    # So we use "DB START 0" to get the root-ExceptionGroup
                    # and "DB START" to automatically select the traceback.
                    if user_selected_comb == (0,):
                        user_selected_comb = ()
                    if user_selected_comb not in exc_combs:
                        user_selected_comb = None
                except Exception:
                    user_selected_comb = None

                if user_selected_comb is None:
                    self.message(
                        "invalid parameter " + repr(arg) + " -- will be ignored"
                    )
                else:
                    selected_comb = user_selected_comb

            if selected_comb is None:
                # automatically select the traceback
                if first_nongroup is not None:
                    selected_comb = first_nongroup
                else:
                    selected_comb = ()

            exc = exc_combs[selected_comb]
            tb = exc.__traceback__

            if user_selected_comb is None:
                cw = 16
                msg_lines = [
                    "There is at least one ExceptionGroup. To select the traceback of a specific group,",
                    "stop the debugger and type the magic command DB START with (nested) group numbers.",
                    "Possible commands for the last traceback are:",
                    "db stop".ljust(cw) + "<-- stops the debugger",
                    "db traceback".ljust(cw) + "<-- prints the last traceback again",
                    "db start".ljust(cw)
                    + "<-- this will automatically choose the first non-ExceptionGroup",
                ]
                for comb in sorted(exc_combs.keys()):
                    comment = ""
                    if comb == selected_comb:
                        if selected_comb == first_nongroup:
                            comment = "<-- currently used (automatically chosen -- first non-exc-group)"
                        else:
                            comment = (
                                "<-- currently used (automatically chosen -- fallback)"
                            )

                    if comb == ():
                        comb_string = "0"
                        if comment == "":
                            comment = "<-- zero means: explicitly choose the root ExceptionGroup"
                    else:
                        comb_string = " ".join([str(a) for a in comb])
                    msg_lines.append(("db start " + comb_string).ljust(cw) + comment)
                self.message("\n".join(msg_lines))
        return exc, tb

    def do_start(self, arg):
        """Start postmortem debugging from the last uncaught exception."""

        exc = sys.last_value

        # Python 2.7 exceptions have no __traceback__ attribute
        tb = sys.last_traceback

        # Since the introduction of ExceptionGroup, there could be multiple tracebacks
        # of interest. A specific one can be selected using the "arg" parameter.
        exc, tb = self._get_selected_exception(exc, tb, arg)

        # Interact, or not
        if self._debugmode:
            self.message("Already in debug mode.")
        elif tb:
            self.interaction(None, tb, pm=True, pm_exc=exc)
        else:
            self.message("No debug information available.")

    def do_frame(self, arg):
        """Go to the i'th frame in the stack."""
        interpreter = sys._pyzoInterpreter

        if not self._debugmode:
            self.message("Not in debug mode.")
        else:
            # Set frame index
            interpreter._dbFrameIndex = int(arg)
            if interpreter._dbFrameIndex < 1:
                interpreter._dbFrameIndex = 1
            elif interpreter._dbFrameIndex > len(interpreter._dbFrames):
                interpreter._dbFrameIndex = len(interpreter._dbFrames)
            # Set name and locals
            frame = interpreter._dbFrames[interpreter._dbFrameIndex - 1]
            interpreter._dbFrameName = frame.f_code.co_name
            interpreter.locals = frame.f_locals
            interpreter.globals = frame.f_globals
            self.writestatus()

    def do_up(self, arg):
        """Go one frame up the stack."""
        interpreter = sys._pyzoInterpreter

        if not self._debugmode:
            self.message("Not in debug mode.")
        else:
            # Decrease frame index
            interpreter._dbFrameIndex -= 1
            if interpreter._dbFrameIndex < 1:
                interpreter._dbFrameIndex = 1
            # Set name and locals
            frame = interpreter._dbFrames[interpreter._dbFrameIndex - 1]
            interpreter._dbFrameName = frame.f_code.co_name
            interpreter.locals = frame.f_locals
            interpreter.globals = frame.f_globals
            self.writestatus()

    def do_down(self, arg):
        """Go one frame down the stack."""
        interpreter = sys._pyzoInterpreter

        if not self._debugmode:
            self.message("Not in debug mode.")
        else:
            # Increase frame index
            interpreter._dbFrameIndex += 1
            if interpreter._dbFrameIndex > len(interpreter._dbFrames):
                interpreter._dbFrameIndex = len(interpreter._dbFrames)
            # Set name and locals
            frame = interpreter._dbFrames[interpreter._dbFrameIndex - 1]
            interpreter._dbFrameName = frame.f_code.co_name
            interpreter.locals = frame.f_locals
            interpreter.globals = frame.f_globals
            self.writestatus()

    def do_stop(self, arg):
        """Stop debugging, terminate process execution."""
        # Can be done both in postmortem and normal debugging
        if not self._debugmode:
            self.message("Not in debug mode.")
        else:
            self.set_quit()
            self.stopinteraction()

    def do_where(self, arg):
        """Print the stack trace and indicate the current frame."""
        interpreter = sys._pyzoInterpreter

        if not self._debugmode:
            self.message("Not in debug mode.")
        else:
            lines = []
            for i in range(len(interpreter._dbFrames)):
                frameIndex = i + 1
                f = interpreter._dbFrames[i]
                # Get fname and lineno, and correct if required
                fname, lineno = f.f_code.co_filename, f.f_lineno
                fname, lineno = interpreter.correctfilenameandlineno(fname, lineno)
                # Build string
                text = 'File "%s", line %i, in %s' % (fname, lineno, f.f_code.co_name)
                if frameIndex == interpreter._dbFrameIndex:
                    lines.append("-> %i: %s" % (frameIndex, text))
                else:
                    lines.append("   %i: %s" % (frameIndex, text))
            lines.append("")
            sys.stdout.write("\n".join(lines))

    def do_continue(self, arg):
        """Continue the program execution."""
        if self._debugmode == 0:
            self.message("Not in debug mode.")
        elif self._debugmode == 1:
            self.message("Cannot use 'continue' in postmortem debug mode.")
        else:
            # the following line is a workaround for a specific problem:
            sys._pyzoInterpreter.apply_breakpoints()

            # run the following code without any breakpoints:
            """
            [breakpoint() for _ in range(1)]
            # [breakpoint()]  # this does not cause any problem
            x = 1
            y = 2  # set a breakpoint in this line once stopped in the upper breakpoint()
            """
            # The breakpoint at the line above would be ignored if:
            # a) the interruption is inside a list comprehension
            # and b) the Python interpreter version is >= 3.0 and < 3.12
            # The problem would also occur when pausing inside a list comprehension.
            # As a workaround we call apply_breakpoints a second time, see above.

            self._last_db_command = "continue"
            self.set_continue()
            self.stopinteraction()

    def do_step(self, arg):
        """Execute the current line, stop ASAP (step into)."""
        if self._debugmode == 0:
            self.message("Not in debug mode.")
        elif self._debugmode == 1:
            self.message("Cannot use 'step' in postmortem debug mode.")
        else:
            self._last_db_command = "step"
            self.set_step()
            self.stopinteraction()

    def do_next(self, arg):
        """Continue execution until the next line (step over)."""
        interpreter = sys._pyzoInterpreter

        if self._debugmode == 0:
            self.message("Not in debug mode.")
        elif self._debugmode == 1:
            self.message("Cannot use 'next' in postmortem debug mode.")
        else:
            frame = interpreter._dbFrames[-1]
            self._last_db_command = "next"
            self.set_next(frame)
            self.stopinteraction()

    def do_return(self, arg):
        """Continue execution until the current function returns (step out)."""
        interpreter = sys._pyzoInterpreter

        if self._debugmode == 0:
            self.message("Not in debug mode.")
        elif self._debugmode == 1:
            self.message("Cannot use 'return' in postmortem debug mode.")
        else:
            self._last_db_command = "return"
            frame = interpreter._dbFrames[-1]
            self.set_return(frame)
            self.stopinteraction()

    def do_jump(self, arg):
        """Jump to a specific line, not executing code in between."""
        interpreter = sys._pyzoInterpreter

        if self._debugmode == 0:
            self.message("Not in debug mode.")
        elif self._debugmode == 1:
            self.message("Cannot use 'jump' in postmortem debug mode.")
        elif interpreter._dbFrameIndex != len(interpreter._dbFrames):
            self.message(
                "Jumping is only possible inside the bottom frame (highest index)."
            )
        else:
            frame = interpreter._dbFrames[interpreter._dbFrameIndex - 1]
            try:
                lineno = int(arg)
                f = interpreter._dbFrames[-1]
                offset = interpreter.correctfilenameandlineno(f.f_code.co_filename, 0)[
                    1
                ]
                frame.f_lineno = lineno - offset
            except ValueError as e:
                self.message("Error DB JUMP: " + str(e))
            self.writestatus()

    def do_events(self, arg):
        """Process GUI events for the integrated GUI toolkit."""
        interpreter = sys._pyzoInterpreter
        interpreter.guiApp.process_events()

    def do_traceback(self, arg):
        """Print the last traceback (again)."""
        interpreter = sys._pyzoInterpreter
        interpreter.showtraceback(True)



import os
import sys

from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo import qt

import pyzo
from pyzo.util import paths


class AboutDialog(QtWidgets.QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle(pyzo.translate("menu dialog", "About Pyzo"))
        self.resize(600, 500)

        # Layout
        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)

        # Create image and title
        im = QtGui.QPixmap(
            os.path.join(pyzo.pyzoDir, "resources", "appicons", "pyzologo64.png")
        )
        imlabel = QtWidgets.QLabel(self)
        imlabel.setPixmap(im)
        textlabel = QtWidgets.QLabel(self)
        textlabel.setText("<h3>Pyzo: the Interactive Editor for Python</h3>")
        #
        titleLayout = QtWidgets.QHBoxLayout()
        titleLayout.addWidget(imlabel, 0)
        titleLayout.addWidget(textlabel, 1)
        #
        layout.addLayout(titleLayout, 0)

        # Create tab bar
        self._tabs = QtWidgets.QTabWidget(self)
        self._tabs.setDocumentMode(True)
        layout.addWidget(self._tabs, 1)

        # Create button box
        self._butBox = QtWidgets.QDialogButtonBox(self)
        self._butBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self._butBox.setStandardButtons(self._butBox.StandardButton.Close)
        layout.addWidget(self._butBox, 0)

        # Signals
        self._butBox.rejected.connect(self.close)

        # Create tabs
        self.createGeneralTab()

    def addTab(self, title, text, rich=True):
        # Create label to show info
        label = QtWidgets.QTextEdit(self)
        label.setLineWrapMode(label.LineWrapMode.WidgetWidth)
        label.setReadOnly(True)
        # Set text
        if rich:
            label.setHtml(text)
        else:
            label.setText(text)
        # Add to tab bar
        self._tabs.addTab(label, title)
        # Done
        return label

    def createGeneralTab(self):
        aboutText = """
        {}<br><br>

        <b>Version info</b><br>
        Pyzo version: <u>{}</u><br>
        Platform: {}<br>
        Python version: {}<br>
        Qt version: {}<br>
        {} version: {}<br>
        <br>

        <b>Pyzo directories</b><br>
        Pyzo source directory: {}<br>
        Pyzo config directory: {}<br>
        Pyzo userdata directory: {}<br>
        <br>

        <b>License</b><br>
        Pyzo is open source, its code is distributed under the 2-Clause BSD license.
        <br><br>

        <b>Contributors</b><br>
        Pyzo is coded with ♥ by Almar Klein and over 30 contributors:<br>
        https://github.com/pyzo/pyzo/graphs/contributors
        <br><br>

        <b>Acknowledgements</b><br>
        Pyzo is written in Python 3 and uses the Qt widget
        toolkit. Pyzo uses code and concepts that are inspired by
        IPython, Pype, and Spyder.
        Pyzo uses a (modified) subset of the silk icon set,
        by Mark James (http://www.famfamfam.com/lab/icons/silk/).
        """
        # Determine if this is PyQt or Pyside
        qtWrapper = qt.API
        qtVersion = qt.QT_VERSION_STR
        qtWrapperVersion = qt.QT_WRAPPER_VERSION_STR
        # Insert information texts
        if paths.is_frozen():
            versionText = pyzo.__version__ + " (binary)"
        else:
            versionText = pyzo.__version__ + " (source)"
        aboutText = aboutText.format(
            "Pyzo - Python to the people!",
            versionText,
            sys.platform,
            sys.version.split(" ")[0],
            qtVersion,
            qtWrapper,
            qtWrapperVersion,
            pyzo.pyzoDir,
            pyzo.appConfigDir,
            pyzo.appDataDir,
        )

        self.addTab("General", aboutText)


if __name__ == "__main__":
    m = AboutDialog(None)
    m.show()



## newline

print('nothing\nspecial\nhere')


## horizontal tab

print('abc\tdef\tghi')
print('1\t2.0\t3.33')


## vertical tab

print('abc\vdefg\vhi')


## backspace

print('1 + 1 = 35', end='')  # stay in this line
print('\b\b2')  # we still can delete characters of the current line and add new ones

## backspace (continued)

# we cannot go further back than the start of the line
print('first line')
print('\b' * 100, end='')  # we cannot delete characters from the line before
print('second line')

## backspace (continued)
import sys

print('1 + 1 = 35', end='')  # stay in this line
print('[unexpected interruption]', end='', file=sys.stderr)
print('\b\b2')  # we cannot delete from a different output stream

## backspace (continued)
import time

# remove characters one by one from the current line
print('first line')
print('12345', end='')
for i in range(10):
    time.sleep(0.3)
    print('\b', end='')
print('finished')


## carriage return

# a '\r' followed by a '\n' is interpreted as a single line break ('\n')
print('first line\r\nsecond line')

# any other character after '\r' means that the current line should be cleared
print('something we want to remove\rcorrected text')

## carriage return (continued)
import time

# the previous examples also work if the text is split into fragments:
print('first line\r', end='')  # the line is already visible here
time.sleep(1.0)
print('\nsecond line')  # now Pyzo knows it was a '\r\n' --> normal line break
time.sleep(1.0)
print('third line\r', end='')  # the line is still visible here
time.sleep(1.0)
print('still third line')  # now Pyzo knows it was a just a'\r' --> remove current (third) line


## format sequence

print('HELLOU\x1b[1;31m\bABCD\x1b[0m')
# note that the '\b' removes the 'U' without damaging the format sequence between

## format sequence (continued)
import time

# If the format sequence is not written in one go, the incomplete sequence will be
# written at first, but then deleted again when the second fragment is processed.
print('abcd' '\x1b[0;1', end='')
time.sleep(1.0)
print(';4;31m' 'hello' '\x1b[39m' 'world' '\x1b[0m' 'efgh')

## format sequence (continued)
import sys

# An interruption by another stream keeps the incomplete format from the first fragment
# but the format is processed correctly anyways:
print('abcd' '\x1b[0;1', end='')
print('HI FROM STDERR', file=sys.stderr)
print(';4;31m' 'hello' '\x1b[39m' 'world' '\x1b[0m' 'efgh')


## formats are stream-specific
import sys

print('normal text to stdout')
print('THIS IS STDERR, in red, and\x1b[0;1;36m now in bold cyan', file=sys.stderr)
print('another text to stdout', file=sys.stdout)
print('STDERR HAS KEPT THE FORMAT (still bold cyan)', file=sys.stderr)
print('normal stdout stream is not affected ...')
print('resetting STDERR FORMAT:\x1b[0m DONE', file=sys.stderr)

## formats are stream-specific (continued)
import sys, time, subprocess

# Note that raw streams of other processes, e.g. via subprocess.run(...) share the same
# format for stdout and stderr in the Pyzo shell.

def srun(cmd):
    subprocess.run([sys.executable, '-c', cmd])

print('starting ...')
time.sleep(0.5)
srun("import sys; print('this is stdout via raw stream')")
srun("import sys; print('this is stderr via raw stream', file=sys.stderr)")
srun("import sys; print('changing \x1b[1;32mcolor in stdout raw stream')")
srun("import sys; print('again stdout via raw stream')")
srun("import sys; print('again stderr via raw stream', file=sys.stderr)")
srun("import sys; print('resetting \x1b[0mcolor in stderr raw stream', file=sys.stderr)")
srun("import sys; print('finally stdout via raw stream')")
srun("import sys; print('finally stderr via raw stream', file=sys.stderr)")


## supported text format escape sequences
# this is a subset of https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit

class Fmt:
    reset = 0
    bold = 1
    light = 2
    italic = 3
    underline = 4
    no_bold_light = 22
    no_italic = 23
    no_underline = 24
    black = 30
    red = 31
    green = 32
    yellow = 33
    blue = 34
    magenta = 35
    cyan = 36
    white = 37

def build_fmt(*fmt_values):
    return '\x1b[' + ';'.join([str(v) for v in fmt_values]) + 'm'

for k, v in Fmt.__dict__.items():
    if not k.startswith('_'):
        print('\x1b[{}m{}\x1b[0m'.format(v, k))

print()

normal = build_fmt(Fmt.reset)
underline = build_fmt(Fmt.underline)
bold_red = build_fmt(Fmt.bold, Fmt.red)
blue_italic = build_fmt(Fmt.blue, Fmt.italic)
yellow = build_fmt(Fmt.yellow)

print(f'Hello {underline}world {bold_red}via {blue_italic}Py{yellow}thon{normal}!')


## resetting only the text color via escape code 39

print(
    '\x1b[3m'
    'italic'  # --> default color, italic
    '  '
    '\x1b[31m'
    'italic red'  # --> as before, but with red text
    '  '
    '\x1b[39m'  # --> reset text color, but keep italic
    'normal color'
    '  '
    '\x1b[0m'  # --> reset whole format (use color, and remove italic)
    'normal'
)


## progress bar example

import time

for i, p in enumerate(range(0, 100+1, 2)):
    bar = '=' * (i + 1) + ' ' * (50 - i)
    colored_bar = '\x1b[{}m'.format(32 if p == 100 else 31) + bar + '\x1b[0m'
    c = '-\\|/'[i % 4]
    if p == 100:
        c = 'finished'
    print('\rprogress: [' + colored_bar + ']', p, '%  ', c, end='')
    time.sleep(0.1)
print()


## progress bar using package tqdm

import time, tqdm

for i in tqdm.tqdm(range(10)):
    time.sleep(0.5)


## automatic line splitting

# To improve performance with text wrapping in the shell, very long lines are split
# into lines of 80 characters.

print('A' * 1050, end='')  # lines longer than 1024 chars are split
print('B' * 500)  # still split, because this is in the same line as the 'A's
print('c' * 500)  # still split, because we had long lines before
print('d' * 50)  # line shorter than 81 chars (and a '\n') stops splitting long lines
print('e' * 500)  # not split, because shorter than 1024 chars



import sys
import os
import code
import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo.core.shell import BaseShell
from pyzo.core.pyzoLogging import splitConsole

tool_name = pyzo.translate("pyzoLogger", "Logger")
tool_summary = "Logs messages, warnings and errors within Pyzo."


class PyzoLogger(QtWidgets.QWidget):
    """PyzoLogger

    The main widget for this tool.

    """

    def __init__(self, parent):
        super().__init__(parent)

        # logger widget
        self._logger_shell = PyzoLoggerShell(self)

        # set layout
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addWidget(self._logger_shell, 1)
        # spacing of widgets
        self.layout.setSpacing(0)
        # set margins
        margin = pyzo.config.view.widgetMargin
        self.layout.setContentsMargins(margin, margin, margin, margin)
        self.setLayout(self.layout)

    def updateZoom(self):
        self._logger_shell.setZoom(pyzo.config.view.zoom)

    def updateFont(self):
        self._logger_shell.setFont(pyzo.config.view.fontname)


class PyzoLoggerShell(BaseShell):
    """Shell that logs all messages produced by pyzo. It also
    allows to look inside pyzo, which can be handy for debugging
    and developing.
    """

    def __init__(self, parent):
        super().__init__(parent)

        # Set style to Python, or autocompletion does not work
        self.setParser("python3")

        # Change background color to make the logger look different from shell
        f1 = self.getStyleElementFormat("Editor.text")
        fgColor = QtGui.QColor(f1.fore.name())
        bgColor = QtGui.QColor(f1.back.name())
        if pyzo.darkSyntax:
            fgColor = fgColor.lighter(150)
            bgColor = bgColor.lighter(150)
        else:
            fgColor = fgColor.darker(150)
            bgColor = bgColor.darker(110)
        newStyle = "back:{}, fore:{}".format(bgColor.name(), fgColor.name())
        self.setStyle(editor_text=newStyle)

        # Create namespace for logger interpreter
        locals = {"pyzo": pyzo, "sys": sys, "os": os}
        # Include linguist tools
        for name in ["linguist", "lrelease", "lupdate", "lhelp"]:
            locals[name] = getattr(pyzo.util._locale, name)

        # Create interpreter to run code
        self._interpreter = code.InteractiveConsole(locals, "<logger>")

        # Show welcome text
        version = sys.version.split(None, 1)[0]
        moreBanner = "This is the Pyzo logger shell."
        self.write("Python {} on {} - {}\n\n".format(version, sys.platform, moreBanner))
        self.write(str(sys.ps1), 2)

        # Split console
        history = splitConsole(self.write, self.writeErr)
        self.write(history)

    def executeCommand(self, command):
        """Execute the command here!"""
        # Use writeErr rather than sys.stdout.write. This prevents
        # the prompts to be logged by the history. Because if they
        # are, the text does not look good due to missing newlines
        # when loading the history.

        # "Echo" stdin
        self.write(command, 1)
        for line in command.splitlines():
            more = self._interpreter.push(line)
            if more:
                self.write(str(sys.ps2), 2)
            else:
                self.write(str(sys.ps1), 2)

    def writeErr(self, msg):
        """This is what the logger uses to write errors."""
        self.write(msg, 0, "#f00")

    # Note that I did not (yet) implement calltips

    def processAutoComp(self, aco):
        """Processes an autocomp request using an AutoCompObject instance."""

        # Try using buffer first
        if aco.tryUsingBuffer():
            return

        # Include builtins?
        if not aco.name:
            command = "__builtins__.keys()"
            try:
                names = eval(command, {}, self._interpreter.locals)
                aco.addNames(names)
            except Exception:
                pass

        # Query list of names
        command = "dir({})".format(aco.name)
        try:
            names = eval(command, {}, self._interpreter.locals)
            aco.addNames(names)
        except Exception:
            pass

        # Done
        aco.finish()



"""Module baseTextCtrl

Defines the base text control to be inherited by the shell and editor
classes. Implements styling, introspection and a bit of other stuff that
is common for both shells and editors.

"""

import pyzo
import os
import sys
import time
from pyzo.core.pyzoLogging import print
import pyzo.codeeditor.parsers.tokens as Tokens
from pyzo.codeeditor import CodeEditor

from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo.util import CalmedFunc

ismacos = sys.platform.startswith("darwin")


def normalizePath(path):
    """Normalize the path given.
    All slashes will be made the same (and doubles removed)
    The real case as stored on the file system is recovered.
    """
    path = os.path.abspath(path)  # make sure it is defined from the drive up
    path = os.path.normpath(path)
    return path


def isFormatStringToken(token):
    """returns True if a StringToken is a format string literal, e.g. f'{1 + 2}'"""
    if not isinstance(token, Tokens.StringToken):
        return False
    s = str(token)
    return "f" in s[: s.index(s[-1])].lower()


def isSimpleExpression(tokens):
    """returns True if the list of tokens is a simple expression

    a "simple expression" is defined as follows:
    - first token is either:
        - an identifier
        - a string token, but not a format string token
    - further tokens are only identifier tokens, with a dot as a separator

    examples for simple expressions, as stringified tokens:
        myvar123
        myvar.myattribute
        "just a string"
        '\n'.join
        b"xyz".foo.bar.z

    Simple expressions can be mostly evaluated without side effects, at least when
    properties and attribute access dunder methods are assumed to not modify any data.
    """
    if len(tokens) == 0:
        return True
    if len(tokens) % 2 != 1:
        return False
    for i, t in enumerate(tokens):
        if i == 0:
            if isinstance(t, Tokens.StringToken) and not isFormatStringToken(t):
                continue
        if i % 2 == 0 and not isinstance(t, Tokens.IdentifierToken):
            return False
        elif i % 2 == 1 and str(t) != ".":
            return False
    return True


def skipNestedTokensRightToLeft(tokens, indClosing):
    """finds the left end of the nested parens expression given the right end

    Parens can be round parentheses, square brackets and curly braces.
    "indClosing" is the index for list "tokens" where the closing paren is located.
    Tokens are examinded from right to left, only testing for correct matching of
    nested parens.

    return value:
        the index of the corresponding token that matches the closing paren
        or None if the paren could not be found or nested parens do not match
    """
    stack = []
    openingParensDict = {")": "(", "]": "[", "}": "{"}
    for i in range(indClosing, -1, -1):
        t = tokens[i]
        s = str(t)
        if s in ")]}":
            stack.append(s)
            continue
        if s in "([{":
            if len(stack) == 0 or openingParensDict[stack.pop()] != s:
                return None  # parens do not match
            if len(stack) == 0:
                break  # finished successfully
    else:
        return None  # matching paren not found
    indOpening = i
    return indOpening


def getExpressionTokensRightToLeft(tokens, indRightStart):
    """finds the start of an "expression" in a list of tokens, starting at indRightStart

    In this case, an "expression" is a list of one or more tokens that are a subsequence
    of the given list "tokens", with the last token at given index "indRightStart", that
    is the rightmost part of the "expression".

    return value:
        the index of the corresponding token where the "expression" ends on the left
        or None if no expression could be found or nested parens do not match

    An "expression" consists of a sequence of token groups.
    A token group can be:
        - an identifier
        - a dot
        - a string token
        - a (nested) parens group of tokens
    A dot is only allowed on the left of an identifier (i.e. on the left of the attribute).
    A curly parens group can only be the first token group (on the left of the expression).
    Dots can not be placed at the very left or very right of the expression, only in between.

    examples for "expressions", as stringified tokens:
        b"x".hex()
        'abc'.join([])
        (1, 2)[0].bit_length().__dir__()[0].upper()
        f'{3 + 4}'.join
        myvar
        (1, 2)[0].bit_length
        [11, 33].count
        [11, 33].count(5)
        (2).bit_length()
        {'a': 1}[b]
        {'a': 1}.get(x)

    Evaluating the values of such "expressions" during introspection could trigger a
    series of calculations and modify data during introspection, for example when
    consuming values of an iterator, or when inspecting code like "mylist.pop()['abc']".
    """
    i = indRightStart + 1
    exprTokens = []  # this will grow to the right when when decreasing the token index (reversed order)
    containsCurly = False
    while i > 0:
        i -= 1
        t = tokens[i]
        s = str(t)
        if s in ")]}":
            indOpening = skipNestedTokensRightToLeft(tokens, i)
            if indOpening is None:
                return None  # parens do not match
            if s == "}":
                if containsCurly:
                    return None  # curly braces encountered more than once
                containsCurly = True
            exprTokens.extend(tokens[indOpening : i + 1][::-1])
            i = indOpening
        elif s == ".":
            if len(exprTokens) == 0 or not isinstance(
                exprTokens[-1], Tokens.IdentifierToken
            ):
                return None  # dot is not on the left of an attribute
            exprTokens.append(t)
        elif isinstance(t, Tokens.IdentifierToken):
            if len(exprTokens) > 0 and isinstance(
                exprTokens[-1], (Tokens.IdentifierToken, Tokens.StringToken)
            ):
                return None  # an identifier cannot be directly on the left of a string literal or another identifier
            exprTokens.append(t)
        elif isinstance(t, Tokens.StringToken):
            if len(exprTokens) > 0 and isinstance(
                exprTokens[-1], Tokens.IdentifierToken
            ):
                return None  # a string literal cannot be directly on the left of an identifier
            exprTokens.append(t)
        else:
            indLeftEnd = i + 1
            break
    else:
        indLeftEnd = 0

    if len(exprTokens) == 0:
        return None  # no expression found

    if containsCurly and str(exprTokens[-1]) != "{":
        return None  # curly braces can only occur at the very left

    if str(exprTokens[-1]) == ".":
        return None  # a dot cannot occur at the very left

    return indLeftEnd


def parseLine_autocomplete(tokens):
    """gets the name part and the attribute part of an expression in tokens

    The tokens of the expression are located at the very right in the token list.

    return value:
        tuple (nameTokens, needleToken) where "needleToken" is the attribute part
            needleToken will be None if the last token is "."
            nameTokens will be an empty list if there is no dot in the expression
        or tuple (None, None) if no valid expression could be found

    examples with stringified input tokens and result tokens:
        abc = (myfunc(x + ((y + 2) + otherfunc(1, 2).
            name: otherfunc(1, 2)
            needle: None

        abc = (myfunc(x + ((y + 2) + otherfunc(1, 2).xy
            name: otherfunc(1, 2)
            needle: xy

        myvar
            name: [empty list]
            needle: myvar

        eat = food.fruit.ban
            name: food.fruit
            needle: ban

        >>> enu
            name: [empty list]
            needle: enumerate
    """
    retvalInvalid = (None, None)
    if len(tokens) == 0:
        return retvalInvalid

    needleToken = tokens[-1]
    if str(needleToken) == ".":
        needleToken = None
        if len(tokens) == 1:
            return retvalInvalid
        indRightStart = len(tokens) - 1 - 1
    else:
        if not isinstance(needleToken, Tokens.IdentifierToken):
            return retvalInvalid
        if len(tokens) == 1 or str(tokens[-2]) != ".":
            nameTokens = []
            return nameTokens, needleToken
        indRightStart = len(tokens) - 1 - 2

    indLeftEnd = getExpressionTokensRightToLeft(tokens, indRightStart)
    if indLeftEnd is None:
        return retvalInvalid
    nameTokens = tokens[indLeftEnd : indRightStart + 1]
    return nameTokens, needleToken


def parseLine_signature(tokens, paren="("):
    """gets the expression before the open paren on the right side in a token list

    Processes the tokens from right to left till it finds an open paren that is
    preceded by an "expression" consisting of one or more tokens. Also gets the index
    of the token that is the open paren.
    Paren groups that are encountered while moving from right to left must be properly
    matched by the opposite paren token, also for nested paren groups.
    Parens that are only open to the right and not preceded by an expression, are ignored.

    return value:
        tuple (fullNameTokens, indOfOpenParen)
        or tuple (None, None) if not found or invalid parens groups encountered

    examples with stringified input tokens and result tokens:
        abc = (myfunc(x + ((y + 2) + otherfunc(1, 2) -
            fullName: myfunc
            indOfOpenParen: 4

        abc = (getfunc('aa')(x + ((y + 2) + otherfunc(1, 2) -
            fullName: getfunc('aa')
            indOfOpenParen: 7
    """
    retvalInvalid = (None, None)
    i = len(tokens)
    while i > 0:
        i -= 1
        t = tokens[i]
        s = str(t)
        if s in ")]}":
            indOpening = skipNestedTokensRightToLeft(tokens, i)
            if indOpening is None:
                return retvalInvalid  # nested paren group is not complete
            i = indOpening
            continue
        if s == paren:
            indLeftEnd = getExpressionTokensRightToLeft(tokens, i - 1)
            if indLeftEnd is not None:
                fullNameTokens, indParenToken = tokens[indLeftEnd:i], i
                return fullNameTokens, indParenToken

    return retvalInvalid  # no open paren preceded by an expression found


## examples for running tests:
if False:
    ##
    p = pyzo.shells.getCurrentShell().parser()

    # tokens = p.parseLine("abc = (myfunc(x + ((y + 2) + otherfunc(1, 2).xy")
    # res = pyzo.core.baseTextCtrl.parseLine_autocomplete(tokens)

    tokens = p.parseLine("abc = (getfunc('aa')(x + ((y + 2) + otherfunc(1, 2) -")
    res = pyzo.core.baseTextCtrl.parseLine_signature(tokens)

    print(repr(res))
    if res != (None, None):
        print("".join(str(t) for t in res[0]))
        print(res[1])
##


class BaseTextCtrl(CodeEditor):
    """The base text control class.
    Inherited by the shell class and the Pyzo editor.
    The class implements autocompletion, calltips, and auto-help
    """

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        # Set style/theme
        try:
            theme = pyzo.themes[pyzo.config.settings.theme.lower()]["data"]
            self.setStyle(theme)
            # autocomplete popup theme
            if pyzo.config.view.get("autoComplete_withTheme", False):
                editor_text_theme = theme["editor.text"].split(",")
                popup_background = editor_text_theme[1].split(":")[1]
                popup_text = editor_text_theme[0].split(":")[1]
                autoComplete_theme = "color: {}; background-color:{};".format(
                    popup_text, popup_background
                )
                self.completer().popup().setStyleSheet(autoComplete_theme)
        except Exception as err:
            print("Could not load theme: " + str(err))

        # Set font and zooming
        self.setFont(pyzo.config.view.fontname)
        self.setZoom(pyzo.config.view.zoom)
        self.setShowWhitespace(pyzo.config.view.showWhitespace)
        self.setHighlightMatchingBracket(pyzo.config.view.highlightMatchingBracket)

        # Create calmed functions to avoid unnecessary updates during a quick sequence of
        # keystrokes.
        self._calmedIntrospectNow = CalmedFunc(
            self._introspectNow, pyzo.config.advanced.autoCompDelay
        )

        # Avoid unnecessary introspections while quickly navigating the autocompletion list
        # via arrow keys. For now, we use the same delay parameter as for updating the
        # autocompletion list.
        self._calmedUpdateHelpFromAutocomplete = CalmedFunc(
            self.updateHelpFromAutocomplete, pyzo.config.advanced.autoCompDelay
        )

        # For buffering autocompletion and calltip info
        self._callTipBuffer_name = ""
        self._callTipBuffer_intermediateResultName = None
        self._callTipBuffer_time = 0
        self._callTipBuffer_result = ""
        self._autoCompBuffer_name = ""
        self._autoCompBuffer_intermediateResultName = None
        self._autoCompBuffer_time = 0
        self._autoCompBuffer_result = []

        self.setAutoCompletionAcceptKeysFromStr(
            pyzo.config.settings.autoComplete_acceptKeys
        )

        self.completer().highlighted.connect(self._onAutocompleteSelectionChanged)
        self.setIndentUsingSpaces(pyzo.config.settings.defaultIndentUsingSpaces)
        self.setIndentWidth(pyzo.config.settings.defaultIndentWidth)
        self.setAutocompletPopupSize(*pyzo.config.view.autoComplete_popupSize)
        self.setAutocompleteCaseSensitive(
            pyzo.config.settings.autoComplete_caseSensitive
        )
        self.setAutocompleteMinChars(pyzo.config.settings.autoComplete_minChars)
        self.setAutoClose_Quotes(pyzo.config.settings.autoClose_Quotes)
        self.setAutoClose_Brackets(pyzo.config.settings.autoClose_Brackets)
        self.setAutocompleteFinishedCallback(self._onAutocompleteFinished)
        self.setCalltipFinishedCallback(lambda: self.restoreHelp("calltip"))

    def _onAutocompleteFinished(self, completionPerformed):
        # overridden in the editor (PyzoEditor class)
        self.restoreHelp("autocomplete")

    def _onAutocompleteSelectionChanged(self, name):
        # overridden in the editor (PyzoEditor class)

        # Update the delay just in case the user changed the value.
        # Otherwise the user would have to restart Pyzo for the changes to take effect.
        self._calmedUpdateHelpFromAutocomplete.set_delay(
            pyzo.config.advanced.autoCompDelay
        )

        self._calmedUpdateHelpFromAutocomplete(name)

    def setAutoCompletionAcceptKeysFromStr(self, keys):
        """Set the keys that can accept an autocompletion from a comma delimited string."""
        # Set autocomp accept key to default if necessary.
        # We force it to be string (see issue 134)
        if not isinstance(keys, str):
            keys = "Tab"
        # Split
        keys = keys.replace(",", " ").split(" ")
        keys = [key for key in keys if key]
        # Set autocomp accept keys
        qtKeys = []
        for key in keys:
            if len(key) > 1:
                key = "Key_" + key[0].upper() + key[1:].lower()
                qtkey = getattr(QtCore.Qt.Key, key, None)
            else:
                qtkey = ord(key)
            if qtkey:
                qtKeys.append(qtkey)

        if QtCore.Qt.Key.Key_Enter in qtKeys and QtCore.Qt.Key.Key_Return not in qtKeys:
            qtKeys.append(QtCore.Qt.Key.Key_Return)
        self.setAutoCompletionAcceptKeys(*qtKeys)

    def getTokensUpToCursor(self, cursor):
        # In order to find the tokens, we need the userState from the highlighter
        if cursor.block().previous().isValid():
            previousState = cursor.block().previous().userState()
        else:
            previousState = 0

        text = cursor.block().text()[: cursor.positionInBlock()]

        return (
            text,
            [
                t for t in self.parser().parseLine(text, previousState) if t.isToken
            ],  # filter to remove BlockStates
        )

    def introspect(self, tryAutoComp=False, delay=True, advanced=False):
        """The starting point for introspection (autocompletion and calltip).
        It will always try to produce a calltip. If tryAutoComp is True,
        will also try to produce an autocompletion list (which, on success,
        will hide the calltip). If advanced is True, introspection will also
        evaluate more complexe expressions that contain function calls and/or
        indexing operations, and this will create cache variables in the
        shell's current scope.

        This method will obtain the line and (re)start a timer that will
        call _introspectNow() after a short while. This way, if the
        user types a lot of characters, there is not a stream of useless
        introspection attempts; the introspection is only really started
        after he stops typing for, say 0.1 or 0.5 seconds (depending on
        pyzo.config.autoCompDelay).

        The method _introspectNow() will parse the line to obtain
        information required to obtain the autocompletion and signature
        information. Then it calls processCallTip and processAutoComp
        which are implemented in the editor and shell classes.
        """

        # Find the tokens up to the cursor
        cursor = self.textCursor()

        text, tokensUptoCursor = self.getTokensUpToCursor(cursor)

        # TODO: Only proceed if valid python (no need to check for comments/
        # strings, this is done by the processing of the tokens). Check for python style

        # Is the char valid for auto completion?
        if tryAutoComp:
            if not text:
                self.autocompleteCancel()
                tryAutoComp = False

        cursor.setKeepPositionOnInsert(True)

        # Update the delay just in case the user changed the value.
        # Otherwise the user would have to restart Pyzo for the changes to take effect.
        self._calmedIntrospectNow.set_delay(pyzo.config.advanced.autoCompDelay)

        if not delay:
            self._calmedIntrospectNow.clear_pending()
        self._calmedIntrospectNow(tokensUptoCursor, cursor, tryAutoComp, advanced)

    def _introspectNow(self, tokensUptoCursor, cursor, tryAutoComp, advanced):
        """This method is called a short while after introspect()
        by the timer. It parses the line and calls the specific methods
        to process the callTip and autoComp.
        """
        tokens = tokensUptoCursor

        if pyzo.config.settings.autoCallTip:
            # Parse the line, to get the name of the function we should calltip
            # if the name is empty/None, we should not show a signature
            fullNameTokens, indParenToken = parseLine_signature(tokens)

            useIntermediateResult = False
            if advanced:
                self._callTipBuffer_time = 0  # clear buffer

            fullName = ""
            if fullNameTokens is not None:
                # Compose actual name
                fullName = "".join([str(t) for t in fullNameTokens])
                if not isSimpleExpression(fullNameTokens):
                    if advanced or fullName == self._callTipBuffer_name:
                        useIntermediateResult = True
                    else:
                        fullName = ""

            if fullName:
                # Process
                indOfOpenParen = tokens[indParenToken].start
                offset = (
                    cursor.positionInBlock()
                    - indOfOpenParen
                    + len(str(fullNameTokens[-1]))
                )
                cto = CallTipObject(self, fullName, offset, useIntermediateResult)
                self.processCallTip(cto)
            else:
                self.calltipCancel()

        if tryAutoComp and pyzo.config.settings.autoComplete:
            # Parse the line, to see what name or attribute we need to auto-complete
            nameTokens, needleToken = parseLine_autocomplete(tokens)
            keyLookUp = False

            if not nameTokens and not str(needleToken).isidentifier():
                # nameTokens could be None or []
                nameTokens = (
                    None  # force key-auto-completion (for indices, and numeric keys)
                )

            if nameTokens is None:
                # no auto-completion for a name or attribute found --> try key-auto-completion instead
                keyLookUp = True
                nameTokens, indParenToken = parseLine_signature(tokens, paren="[")
                if indParenToken is not None:
                    needleToken = "".join([str(t) for t in tokens[indParenToken + 1 :]])
                else:
                    needleToken = None

            useIntermediateResult = False
            if advanced:
                self._autoCompBuffer_time = 0  # clear buffer

            name = ""
            needle = str(needleToken) if needleToken is not None else ""
            if nameTokens is not None:
                name = "".join([str(t) for t in nameTokens])
                if keyLookUp:
                    name += "["
                if not isSimpleExpression(nameTokens):
                    if advanced or name == self._autoCompBuffer_name:
                        useIntermediateResult = True
                    else:
                        name, needle = "", ""

            if name or needle:
                # Try to do auto completion
                aco = AutoCompObject(self, name, needle, useIntermediateResult)
                self.processAutoComp(aco)

    def processCallTip(self, cto):
        """Overridden in derive class"""
        pass

    def processAutoComp(self, aco):
        """Overridden in derive class"""
        pass

    def helpOnText(self, pos):
        hw = pyzo.toolManager.getTool("pyzointeractivehelp")
        if not hw:
            return
        name = self.textCursor().selectedText().strip()
        if name == "":
            cursor = self.cursorForPosition(pos - self.mapToGlobal(QtCore.QPoint(0, 0)))
            line = cursor.block().text()
            limit = cursor.positionInBlock()
            while limit < len(line) and (
                line[limit].isalnum() or line[limit] in (".", "_")
            ):
                limit += 1
                cursor.movePosition(cursor.MoveOperation.Right)
            _, tokens = self.getTokensUpToCursor(cursor)
            nameTokens, needleToken = parseLine_autocomplete(tokens)
            if nameTokens:
                name = "{}.{}".format(
                    "".join([str(t) for t in nameTokens]), needleToken
                )
            elif needleToken:
                name = str(needleToken)

        if name != "":
            hw.setObjectName(name, True)

    @staticmethod
    def processHelp(name, source, addToHist=False):
        """Show help on the given full object name.
        - called when going up/down in the autocompletion list or showing calltips.
        """
        assert source in ("autocomplete", "calltip")
        # uses parse_autocomplete() to find baseName and objectName

        # Get help tool
        hw = pyzo.toolManager.getTool("pyzointeractivehelp")
        # Get the shell
        shell = pyzo.shells.getCurrentShell()
        # Both should exist
        if not hw or not shell:
            return

        if name:
            hw.helpFromExtension(name, source, addToHist)

    def _getFullAutocompleteName(self, name):
        if self._autoCompBuffer_intermediateResultName is not None:
            s = self._autoCompBuffer_intermediateResultName
            if self._autoCompBuffer_name.endswith("["):
                name = "{}[{}]".format(s, name)  # key auto-completion
            else:
                name = s + "." + name  # attribute auto-completion
        elif self._autoCompBuffer_name:
            s = self._autoCompBuffer_name
            if s.endswith("["):
                name = s + name + "]"  # key auto-completion
            else:
                name = s + "." + name  # attribute auto-completion
        elif not self.completer().completionPrefix():
            # Don't update help if there is no dot or prefix;
            # the choice would be arbitrary
            name = None
        return name

    ## Callbacks
    def updateHelpFromAutocomplete(self, name):
        """A name has been highlighted, show help on that name"""
        fullName = self._getFullAutocompleteName(name)
        if fullName is not None:
            self.processHelp(fullName, "autocomplete")

    @staticmethod
    def restoreHelp(source):
        assert source in ("autocomplete", "calltip")
        hw = pyzo.toolManager.getTool("pyzointeractivehelp")
        if hw:
            hw.helpFromExtension(None, source)

    def event(self, event):
        """Overload main event handler so we can pass Ctrl-C Ctr-V etc, to the main
        window.
        """
        if isinstance(event, QtGui.QKeyEvent):
            # AltModifier maps to the option key on MacOS, and an alternative to control in some cases
            has_control = event.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
            has_alt = event.modifiers() & QtCore.Qt.KeyboardModifier.AltModifier
            # Ignore CTRL+{A-Z} since those keys are handled through the menu
            if (
                (has_control or has_alt)
                and (event.key() >= QtCore.Qt.Key.Key_A)
                and (event.key() <= QtCore.Qt.Key.Key_Z)
            ):
                event.ignore()
                return False

        # Default behavior
        super().event(event)
        return True

    def keyPressEvent(self, event):
        """Receive qt key event.
        From here we will dispatch the event to perform autocompletion
        or other stuff...
        """

        KM = QtCore.Qt.KeyboardModifier

        # Get ordinal key
        ordKey = -1
        if event.text():
            ordKey = ord(event.text()[0])

        # Detect "control". This is awkward:
        # Command Key ⌘ -> ControlModifier -> mostly what Windows uses Control for.
        # Option key ⌥  -> AltModifier     -> For stuff that we cannot use Command key for.
        # Control key ^ -> MetaModifier    -> I guess most uses maps to what Windows uses Alt for.
        has_control_like = event.modifiers() & (
            KM.AltModifier if ismacos else KM.ControlModifier
        )

        # Leave tab events to editor tabs
        # FIXME: this is probably just a workaround for another problem
        # FIXME: ... somewhere in a keyPressEvent in a codeeditor extension
        if (
            ismacos
            and has_control_like
            and event.key()
            in (
                QtCore.Qt.Key.Key_Tab,
                QtCore.Qt.Key.Key_Backtab,
            )
        ):
            event.ignore()
            return

        # Invoke advanced autocomplete/calltips Ctrl+Space key combination?
        if has_control_like and event.key() == QtCore.Qt.Key.Key_Space:
            cursor = self.textCursor()
            if cursor.position() == cursor.anchor():
                text = cursor.block().text()[: cursor.positionInBlock()]
                if text:
                    self.introspect(True, False, advanced=True)
                    return

        # Invoke autocomplete via tab key?
        elif (
            event.modifiers() == KM.NoModifier and event.key() == QtCore.Qt.Key.Key_Tab
        ):
            if pyzo.config.settings.autoComplete and not self.autocompleteActive():
                cursor = self.textCursor()
                if cursor.position() == cursor.anchor():
                    text = cursor.block().text()[: cursor.positionInBlock()]
                    if text and (text[-1] in (Tokens.ALPHANUM + "._")):
                        self.introspect(True, False)
                        return

        super().keyPressEvent(event)

        # Analyse character/key to determine what introspection to fire
        if ordKey:
            if (
                ordKey >= 32 or ordKey == 8  # 8: '\b'
            ) and pyzo.config.settings.autoComplete == 1:
                delay = ordKey != 46  # 46: '.'
                # If a char that allows completion or backspace or dot was pressed
                self.introspect(True, delay)
            elif ordKey >= 32:
                # Printable chars, only calltip
                self.introspect()
        elif event.key() in [QtCore.Qt.Key.Key_Left, QtCore.Qt.Key.Key_Right]:
            self.introspect()


class CallTipObject:
    """Object to help the process of call tips.
    An instance of this class is created for each call tip action.
    """

    def __init__(self, textCtrl, name, offset, useIntermediateResult=False):
        self.textCtrl = textCtrl
        self.name = name
        self.bufferName = name
        self.offset = offset
        self.useIntermediateResult = useIntermediateResult

    def tryUsingBuffer(self):
        """Try performing this callTip using the buffer.

        Returns True on success.
        """
        bufferName = self.textCtrl._callTipBuffer_name
        t = time.time() - self.textCtrl._callTipBuffer_time
        if self.bufferName == bufferName and t < 0:
            self._finish(self.textCtrl._callTipBuffer_result)
            return True
        else:
            return False

    def finish(self, callTipText):
        """Finish the introspection using the given calltipText.

        Will also automatically call setBuffer.
        """
        if self.useIntermediateResult:
            self.setBuffer(callTipText, 1e10)  # almost infinite timeout for the buffer
        else:
            self.setBuffer(callTipText)
        self._finish(callTipText)

    def setBuffer(self, callTipText, timeout=4):
        """Sets the buffer with the provided text."""
        self.textCtrl._callTipBuffer_name = self.bufferName
        self.textCtrl._callTipBuffer_intermediateResultName = (
            "__pyzo__calltip" if self.useIntermediateResult else None
        )
        self.textCtrl._callTipBuffer_time = time.time() + timeout
        self.textCtrl._callTipBuffer_result = callTipText

    def _finish(self, callTipText):
        highlightFunctionName = not self.useIntermediateResult
        # ... because "foo().bar().func(" would only highlight "foo"
        self.textCtrl.calltipShow(self.offset, callTipText, highlightFunctionName)
        BaseTextCtrl.processHelp(self.name, "calltip")


class AutoCompObject:
    """Object to help the process of auto completion.
    An instance of this class is created for each auto completion action.
    """

    def __init__(self, textCtrl, name, needle, useIntermediateResult=False):
        self.textCtrl = textCtrl
        self.bufferName = name  # name to identify with
        self.name = name  # object to find attributes of
        self.needle = needle  # partial name to look for
        self.names = set()  # the names (use a set to prevent duplicates)
        self.useIntermediateResult = useIntermediateResult

    def addNames(self, names):
        """Add a list of names to the collection. Duplicates are removed."""
        self.names.update(names)

    def tryUsingBuffer(self):
        """Try performing this auto-completion using the buffer.

        Returns True on success.
        """
        bufferName = self.textCtrl._autoCompBuffer_name
        t = time.time() - self.textCtrl._autoCompBuffer_time
        if self.bufferName == bufferName and t < 0:
            self._finish(self.textCtrl._autoCompBuffer_result)
            return True
        else:
            return False

    def finish(self):
        """Finish the introspection using the collected names.

        Will automatically call setBuffer.
        """
        # Remember at the object that started this introspection
        # and get sorted names
        timeout = (
            1e10 if self.useIntermediateResult else None
        )  # almost infinite timeout for the buffer
        names = self.setBuffer(self.names, timeout)
        # really finish
        self._finish(names)

    def setBuffer(self, names=None, timeout=None):
        """Sets the buffer with the provided names (or the collected names).

        Also returns a list with the sorted names.
        """
        # Determine timeout
        # Global namespaces change more often than local one, plus when
        # typing a xxx.yyy, the autocompletion buffer changes and is thus
        # automatically refreshed.
        # I've once encountered a wrong autocomp list on an object, but
        # haven't been able to reproduce it. It was probably some odity.
        if timeout is None:
            if self.bufferName:
                timeout = 5
            else:
                timeout = 1
        # Get names
        if names is None:
            names = self.names
        # Make list and sort
        names = list(names)
        names.sort(key=str.upper)
        # Store
        self.textCtrl._autoCompBuffer_name = self.bufferName
        self.textCtrl._autoCompBuffer_intermediateResultName = (
            "__pyzo__autocomp" if self.useIntermediateResult else None
        )
        self.textCtrl._autoCompBuffer_time = time.time() + timeout
        self.textCtrl._autoCompBuffer_result = names
        # Return sorted list
        return names

    def _finish(self, names):
        # Show completion list if required.
        self.textCtrl.autocompleteShow(len(self.needle), names, self.name != "")

    def nameInImportNames(self, importNames):
        """Test whether the name, or a base part of it is present in the
        given list of names. Returns the (part of) the name that's in
        the list, or None otherwise.
        """
        baseName = self.name
        while baseName not in importNames:
            if "." in baseName:
                baseName = baseName.rsplit(".", 1)[0]
            else:
                baseName = None
                break
        return baseName


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    win = BaseTextCtrl(None)
    #     win.setStyle('.py')
    tmp = "foo(bar)\nfor bar in range(5):\n  print bar\n"
    tmp += "\nclass aap:\n  def monkey(self):\n    pass\n\n"
    tmp += "a\u20acb\n"
    win.setPlainText(tmp)
    win.show()
    app.exec()
